from paradox.interfaces.mqtt.entities.abstract_entity import AbstractEntity
from paradox.lib.utils import sanitize_key


class AbstractStatusBinarySensor(AbstractEntity):
    def __init__(self, entity, property: str, device, availability_topic: str):
        super().__init__(device, availability_topic)

        self.label = entity.get("label", entity["key"].replace("_", " "))
        self.property = property

        self.key = sanitize_key(entity["key"])

        self.hass_entity_type = "binary_sensor"

    def serialize(self):
        config = super().serialize()
        config.update(dict(
            payload_on="True",
            payload_off="False",
        ))
        return config


class PartitionBinarySensor(AbstractStatusBinarySensor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.pai_entity_type = "partition"


class ZoneStatusBinarySensor(AbstractStatusBinarySensor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.pai_entity_type = "zone"

    def serialize(self):
        config = super().serialize()
        if self.property == 'open':
            config['device_class'] = 'motion'

        return config


class SystemBinarySensor(AbstractStatusBinarySensor):
    def __init__(self, key: str, property: str, device, availability_topic: str):
        super().__init__({"key": key}, property, device, availability_topic)

        self.pai_entity_type = "system"



#!/usr/bin/env python3

__author__ = "João Paulo Barraca, Jevgeni Kiski"
__copyright__ = "Copyright 2018-2019, João Paulo Barraca"
__credits__ = ["Tihomir Heidelberg", "Louis Rossouw"]
__license__ = "EPL"
__maintainer__ = "João Paulo Barraca"
__email__ = "jpbarraca@gmail.com"
__status__ = "Beta"

import argparse
import asyncio
import logging
import os
import sys

from paradox.config import config as cfg
from paradox.lib import help
from paradox.lib.encodings import register_encodings
from paradox.paradox import Paradox

if sys.version_info < (3, 6,):
    print(
        "You are using Python %s.%s, but PAI requires at least Python 3.6"
        % (sys.version_info[0], sys.version_info[1])
    )
    sys.exit(-1)

logger = logging.getLogger("PAI").getChild(__name__)


async def dump_memory(file, memory):
    alarm = Paradox()
    if not await alarm.connect():
        logger.error("Failed to connect")

    await alarm.dump_memory(file, memory)
    await alarm.disconnect()


def main():
    parser = argparse.ArgumentParser()
    types = parser.add_mutually_exclusive_group(required=True)
    types.add_argument(
        "-r", "--ram", dest="type", action="store_const", help="Dump RAM", const="ram"
    )
    types.add_argument(
        "-e",
        "--eeprom",
        dest="type",
        action="store_const",
        help="Dump EEPROM",
        const="eeprom",
    )
    parser.add_argument(
        "-f",
        "--file",
        type=argparse.FileType("wb"),
        default=sys.stdout,
        help="Dump to file. Default stdout",
    )
    parser.add_argument(
        "-c",
        "--config",
        type=str,
        default=None,
        help="specify path to an alternative configuration file",
    )

    args = parser.parse_args()

    if "config" in args and args.config is not None:
        config_file = os.path.abspath(args.config)
        cfg.load(config_file)
    else:
        cfg.load()

    # Registering additional encodings
    register_encodings()

    loop = asyncio.get_event_loop()
    loop.run_until_complete(dump_memory(args.file, args.type))


if __name__ == "__main__":
    try:
        main()
    except ImportError as error:
        help.import_error_help(error)



from paradox.interfaces.mqtt.entities.abstract_entity import AbstractControllableEntity
from paradox.interfaces.mqtt.entities.device import Device
from paradox.lib.utils import sanitize_key


class AlarmControlPanel(AbstractControllableEntity):
    def __init__(self, partition: dict, device: Device, availability_topic: str, code: str = None):
        super(AlarmControlPanel, self).__init__(device, availability_topic)

        self.key = sanitize_key(partition["key"])
        self.label = partition["label"]
        self.property = "current_state"

        self.hass_entity_type = "alarm_control_panel"
        self.pai_entity_type = "partition"

        self.code = code

    def serialize(self):
        config = super().serialize()
        config.update(dict(
            payload_disarm="disarm",
            payload_arm_home="arm_stay",
            payload_arm_away="arm",
            payload_arm_night="arm_sleep"
        ))
        if self.code is not None:
            config['code']=self.code
        else:
            config.update(dict(
                code_arm_required=False,
                code_disarm_required=False,
                code_trigger_required=False
            ))
        return config

    @property
    def entity_id(self):
        return f"{self.pai_entity_type}_{self.key.lower()}"

    @property
    def entity_name(self):
        return f"{self.pai_entity_type.title()} {self.label}"


from paradox.config import config as cfg, get_limits_for_type

def test_get_limits_for_type_for_auto(mocker):
    mocker.patch.multiple(
        cfg,
        LIMITS={
            "partition": 'auto',
        }
    )


    assert get_limits_for_type('partition') == None
    assert get_limits_for_type('partition', []) == []


def test_get_limits_for_type_for_specified_string_range(mocker):
    mocker.patch.multiple(
        cfg,
        LIMITS={
            "partition": '1-4',
        }
    )


    assert get_limits_for_type('partition') == [1,2,3,4]
    assert get_limits_for_type('partition', []) == [1,2,3,4]


def test_get_limits_for_type_for_specified_string_list(mocker):
    mocker.patch.multiple(
        cfg,
        LIMITS={
            "partition": '1,4',
        }
    )


    assert get_limits_for_type('partition') == [1,4]
    assert get_limits_for_type('partition', []) == [1,4]


def test_get_limits_for_type_for_specified_py_range(mocker):
    mocker.patch.multiple(
        cfg,
        LIMITS={
            "partition": range(1,5),
        }
    )


    assert get_limits_for_type('partition') == [1,2,3,4]
    assert get_limits_for_type('partition', []) == [1,2,3,4]


def test_get_limits_for_type_for_specified_py_list(mocker):
    mocker.patch.multiple(
        cfg,
        LIMITS={
            "partition": [1,2],
        }
    )


    assert get_limits_for_type('partition') == [1,2]
    assert get_limits_for_type('partition', []) == [1,2]


def test_get_limits_for_type_for_specified_py_None(mocker):
    mocker.patch.multiple(
        cfg,
        LIMITS={}
    )


    assert get_limits_for_type('partition') == None
    assert get_limits_for_type('partition', []) == []


def test_get_limits_for_type_for_specified_empty_string(mocker):
    mocker.patch.multiple(
        cfg,
        LIMITS={
            'partition': ''
        }
    )


    assert get_limits_for_type('partition') == []
    assert get_limits_for_type('partition', [1,2]) == []


import asyncio
import binascii
from unittest import mock

from construct import Container
import pytest

from paradox.hardware.evo.parsers import LiveEvent, ReadEEPROMResponse
from paradox.lib.async_message_manager import AsyncMessageManager
from paradox.lib.handlers import PersistentHandler


class EventMessageHandler(PersistentHandler):
    def can_handle(self, message):
        return message.fields.value.po.command == 0xE


def print_beer(m):
    print("beer")


@pytest.mark.asyncio
async def test_event_handler():
    eh = EventMessageHandler(print_beer)

    mh = AsyncMessageManager()
    mh.register_handler(eh)

    assert 1 == len(mh.handler_registry)

    payload = b"\xe2\xff\xad\x06\x14\x13\x01\x04\x0e\x10\x00\x01\x05\x00\x00\x00\x00\x00\x02Living room     \x00\xcc"

    message = LiveEvent.parse(payload)

    await mh.schedule_message_handling(message)

    assert 1 == len(mh.handler_registry)


@pytest.mark.asyncio
async def test_event_handler_failure():
    # eeprom_request_bin = binascii.unhexlify('500800009f004037')
    eeprom_response_bin = binascii.unhexlify(
        "524700009f0041133e001e0e0400000000060a0000000000000000000000000000000000000000000000000000000000000000000000000000000000000000121510010705004e85"
    )

    eh = EventMessageHandler(print_beer)
    eh.handle = mock.MagicMock()

    mh = AsyncMessageManager()
    mh.register_handler(eh)

    assert 1 == len(mh.handler_registry)

    message = ReadEEPROMResponse.parse(eeprom_response_bin)

    coro = mh.schedule_message_handling(message)
    result = await coro
    assert (
        result is None
    )  # failed to parse response message return None. Maybe needs to throw something.

    assert 1 == len(mh.handler_registry)
    eh.handle.assert_not_called()


async def test_handler_two_messages():
    def event_handler(message):
        print("event")

    async def get_eeprom_result(mhm):
        return await mhm.wait_for_message(lambda m: m.fields.value.po.command == 0x5)

    event_response_bin = b"\xe2\xff\xad\x06\x14\x13\x01\x04\x0e\x10\x00\x01\x05\x00\x00\x00\x00\x00\x02Living room     \x00\xcc"

    eeprom_response_bin = binascii.unhexlify(
        "524700009f0041133e001e0e0400000000060a0000000000000000000000000000000000000000000000000000000000000000000000000000000000000000121510010705004e85"
    )

    mh = AsyncMessageManager()

    event_handler = EventMessageHandler(event_handler)
    mh.register_handler(event_handler)

    # running
    task_handle_wait = asyncio.create_task(asyncio.sleep(0.1))
    task_get_eeprom = asyncio.create_task(get_eeprom_result(mh))
    task_handle_event1 = mh.schedule_message_handling(
        LiveEvent.parse(event_response_bin)
    )
    mh.schedule_message_handling(ReadEEPROMResponse.parse(eeprom_response_bin))
    mh.schedule_message_handling(LiveEvent.parse(event_response_bin))

    await asyncio.gather(task_handle_wait, task_get_eeprom)

    assert 1 == len(mh.handler_registry)

    assert task_handle_event1.done()
    assert isinstance(task_get_eeprom.result(), Container)

    assert 1 == len(mh.handler_registry)


@pytest.mark.asyncio
async def test_handler_timeout():
    def event_handler(message):
        print("event received")

    async def get_eeprom_result(mhm):
        return await mhm.wait_for_message(
            lambda m: m.fields.value.po.command == 0x5, timeout=0.1
        )

    async def post_eeprom_message(mhm):
        await asyncio.sleep(0.2)

        eeprom_response_bin = binascii.unhexlify(
            "524700009f0041133e001e0e0400000000060a0000000000000000000000000000000000000000000000000000000000000000000000000000000000000000121510010705004e85"
        )

        return await mhm.schedule_message_handling(
            ReadEEPROMResponse.parse(eeprom_response_bin)
        )

    mh = AsyncMessageManager()

    # running
    task_get_eeprom = asyncio.create_task(get_eeprom_result(mh))
    asyncio.create_task(post_eeprom_message(mh))

    assert 0 == len(mh.handler_registry)

    with pytest.raises(asyncio.TimeoutError):
        await task_get_eeprom

    assert 0 == len(mh.handler_registry)

    # Also test EventMessageHandler
    event_handler_instance = EventMessageHandler(event_handler)
    mh.register_handler(event_handler_instance)

    event_response_bin = b"\xe2\xff\xad\x06\x14\x13\x01\x04\x0e\x10\x00\x01\x05\x00\x00\x00\x00\x00\x02Living room     \x00\xcc"
    task_handle_event1 = mh.schedule_message_handling(
        LiveEvent.parse(event_response_bin)
    )

    assert 1 == len(mh.handler_registry)

    await task_handle_event1

    assert 1 == len(mh.handler_registry)



import asyncio
import binascii
import json
import logging
import time

import requests

from paradox.exceptions import ConnectToSiteFailed, StunSessionRefreshFailed
from paradox.lib import stun

logger = logging.getLogger("PAI").getChild(__name__)


class StunSession:
    def __init__(self, site_id, email, panel_serial):
        self.site_id = site_id
        self.email = email
        self.panel_serial = panel_serial

        self.site_info = None
        self.module = None
        self.stun_control = None
        self.stun_tunnel = None

        self.connection_timestamp = 0

    async def connect(self) -> None:
        self.connection_timestamp = 0
        logger.info("Connecting to Site: %s", self.site_id)
        if self.site_info is None:
            self.site_info = await self._get_site_info(
                siteid=self.site_id, email=self.email
            )

        if self.site_info is None:
            raise ConnectToSiteFailed("Unable to get site info")

        logger.debug("Site Info: %s", json.dumps(self.site_info, indent=4))
        self.module = self._select_module()

        if self.module is None:
            self.site_info = None  # Reset state
            raise ConnectToSiteFailed("Unable to find module with desired panel serial")

        xoraddr = binascii.unhexlify(self.module["xoraddr"])

        await self._stun_tcp_change_request()
        await self._stun_tcp_binding_request()
        stun_r = await self._stun_connect(xoraddr)

        self.connection_timestamp = time.time()

        connection_id = stun_r[0]["attr_body"]
        raddr = self.stun_control.sock.getpeername()

        logger.debug("STUN Connection Bind Request")
        self.stun_tunnel = stun.StunClient(host=raddr[0], port=raddr[1])
        self.stun_tunnel.send_connection_bind_request(binascii.unhexlify(connection_id))
        stun_r = self.stun_tunnel.receive_response()
        if stun.is_error(stun_r):
            raise ConnectToSiteFailed(
                f"STUN Connection Bind Request error: {stun.get_error(stun_r)}"
            )

        logger.info("Connected to Site: %s", self.site_id)

    async def _stun_connect(self, xoraddr):
        logger.debug("STUN Connect Request")
        self.stun_control.send_connect_request(xoraddr=xoraddr)
        stun_r = self.stun_control.receive_response()
        if stun.is_error(stun_r):
            raise ConnectToSiteFailed(
                f"STUN Connect Request error: {stun.get_error(stun_r)}"
            )
        return stun_r

    async def _stun_tcp_binding_request(self):
        logger.debug("STUN TCP Binding Request")
        self.stun_control.send_binding_request()
        stun_r = self.stun_control.receive_response()
        if stun.is_error(stun_r):
            raise ConnectToSiteFailed(
                f"STUN TCP Binding Request error: {stun.get_error(stun_r)}"
            )

    async def _stun_tcp_change_request(self):
        stun_host = "turn.paradoxmyhome.com"
        logger.debug("STUN TCP Change Request")
        self.stun_control = stun.StunClient(stun_host)
        self.stun_control.send_tcp_change_request()
        stun_r = self.stun_control.receive_response()
        if stun.is_error(stun_r):
            raise ConnectToSiteFailed(
                f"STUN TCP Change Request error: {stun.get_error(stun_r)}"
            )

    def _select_module(self):
        for site in self.site_info["site"]:
            for module in site["module"]:
                if module.get("xoraddr") is None:
                    continue

                logger.debug(
                    "Found module with panel serial: %s", module["panelSerial"]
                )

                if not self.panel_serial:  # Pick first available
                    return module
                elif module["panelSerial"] == self.panel_serial:
                    return module

    def get_socket(self):
        return self.stun_tunnel.sock

    def refresh_session_if_required(self) -> None:
        if self.site_info is None or self.connection_timestamp == 0:
            return

        # Refresh session if required
        if time.time() - self.connection_timestamp >= 500:
            logger.info("STUN Session Refresh")
            self.stun_control.send_refresh_request()
            stun_r = self.stun_control.receive_response()
            if stun.is_error(stun_r):
                self.connected = False
                raise StunSessionRefreshFailed(
                    f"STUN Session Refresh failed: {stun.get_error(stun_r)}"
                )

            self.connection_timestamp = time.time()

    def close(self):
        self.site_info = None
        self.module = None

        if self.stun_control:
            self.stun_control.close()
            self.stun_control = None
        if self.stun_tunnel:
            self.stun_tunnel.close()
            self.stun_tunnel = None

        self.connection_timestamp = 0

    @staticmethod
    async def _get_site_info(email, siteid):
        logger.info("Getting site info")
        URL = "https://api.insightgoldatpmh.com/v1/site"

        headers = {
            "User-Agent": "Mozilla/3.0 (compatible; Indy Library)",
            "Accept-Encoding": "identity",
            "Accept": "text/html, */*",
        }

        tries = 5
        loop = asyncio.get_event_loop()
        while tries > 0:
            req = await loop.run_in_executor(
                None,
                lambda: requests.get(
                    URL, headers=headers, params={"email": email, "name": siteid}
                ),
            )
            if req.status_code == 200:
                return req.json()

            logger.warning("Unable to get site info. Retrying...")
            tries -= 1
            time.sleep(5)

        return None

    def get_potential_modules(self):
        pass



import binascii
from collections.abc import Mapping

from construct import (Array, BitsInteger, BitsSwapped, BitStruct, Bitwise,
                       Byte, Bytes, ByteSwapped, Checksum, Computed, Const,
                       Default, Embedded, Enum, ExprAdapter,
                       ExprSymmetricAdapter, Flag, Int8ub, Int16ub, Int16ul,
                       Int24ub, Nibble, Padding, RawCopy, Struct, Subconstruct,
                       ValidationError, obj_, this)

from ..common import (CommunicationSourceIDEnum, PacketChecksum, PacketLength,
                      ProductIdEnum, calculate_checksum)
from .adapters import (DateAdapter, DictArray, EnumerationAdapter,
                       EventAdapter, ModuleTroubles, PartitionStatus, PGMFlags,
                       StatusFlags, ZoneFlagBitStruct, ZoneFlags)

LoginConfirmationResponse = Struct(
    "fields"
    / RawCopy(
        Struct(
            "po"
            / BitStruct(
                "command" / Const(0x1, Nibble),
                "status"
                / Struct(
                    "reserved" / Flag,
                    "alarm_reporting_pending" / Flag,
                    "Winload_connected" / Flag,
                    "NeWare_connected" / Flag,
                ),
            ),
            "length" / PacketLength(Int8ub),
            "result"
            / BitStruct(
                "_not_used0" / BitsInteger(3),
                "neware_answer" / Flag,
                "_not_used1" / BitsInteger(4),
            ),
            "callback" / Int16ub,
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)

InitializeCommunication = Struct(
    "fields"
    / RawCopy(
        Struct(
            "po" / Struct("command" / Const(0x00, Int8ub)),
            "module_address" / Default(Int8ub, 0x00),  # (00= panel/module)
            "_not_used0" / Padding(2),
            "product_id" / ProductIdEnum,
            "firmware"
            / Struct("version" / Int8ub, "revision" / Int8ub, "build" / Int8ub),
            "panel_id" / Int16ub,
            "pc_password" / Default(Bytes(2), b"0000"),
            "modem_speed" / Bytes(1),
            "source_method"
            / Default(
                Enum(Int8ub, Winload_Connection=0x00, NEware_Connection=0x55), 0x00
            ),
            "user_code" / Default(Int24ub, 0x000000),
            "serial_number" / Bytes(4),
            "system_options"
            / BitsSwapped(
                BitStruct(
                    Embedded(
                        Struct(  # EVO section data 3030
                            "pgm1_smoke" / Flag,
                            "no_bell_cut_off" / Flag,
                            "daylight_saving_time" / Flag,
                            "shabbat_feature" / Flag,
                            "battery_charge_current" / Flag,
                            "ac_failure_not_displayed_as_trouble" / Flag,
                            "clear_bell_limit_trouble" / Flag,
                            "combus_speed" / Flag,
                        )
                    ),
                    "partitions" / StatusFlags(8),  # EVO section data 3031
                    "siren_output_partition" / StatusFlags(8),  # EVO section data 3032
                    Embedded(
                        Struct(  # EVO section data 3033
                            "multiple_actions_user_menu" / Flag,
                            "user_code_length_flexible" / Flag,
                            "user_code_length_6" / Flag,
                            "power_save_mode" / Flag,
                            "bypass_not_displayed_when_armed" / Flag,
                            "trouble_latch" / Flag,
                            "eol_resistor_on_harwire_zones" / Flag,
                            "atz" / Flag,
                        )
                    ),
                    Embedded(
                        Struct(  # EVO section data 3034
                            "wireless_transmitter_supervision_options" / BitsInteger(2),
                            "generate_supervision_failure_on_bypassed_wireless_zone"
                            / Flag,
                            "restrict_arming_on_wireless_transmitter_supervision_failure"
                            / Flag,
                            "tamper_recognition_options" / BitsInteger(2),
                            "generate_tamper_if_detected_on_bypassed_zone" / Flag,
                            "restrict_arming_on_tamper" / Flag,
                        )
                    ),
                    Embedded(
                        Struct(  # EVO section data 3035
                            "restrict_arming_on_ac_failure" / Flag,
                            "restrict_arming_on_battery_failure" / Flag,
                            "restrict_arming_on_bell_or_aux_failure" / Flag,
                            "restrict_arming_on_tlm_failure" / Flag,
                            "restrict_arming_on_module_troubles" / Flag,
                            "account_number_transmission" / Flag,
                            "transmit_zone_status_on_serial_port" / Flag,
                            "serial_port_baud_rate_57600" / Flag,
                        )
                    ),
                    Embedded(
                        Struct(  # EVO section data 3036
                            "telephone_line_monitoring" / BitsInteger(2),
                            "dialer_reporting" / Flag,
                            "dialing_method" / Flag,
                            "pulse_ratio" / Flag,
                            "busy_tone_detection" / Flag,
                            "switch_to_pulse_dialing" / Flag,
                            "bell_siren_upon_communication_failure" / Flag,
                        )
                    ),
                    Embedded(
                        Struct(  # EVO section data 3037
                            "call_back" / Flag,
                            "automatic_event_buffer_transmission" / Flag,
                            "autotest_report_transmission_options" / BitsInteger(2),
                            "keypad_beep_on_successful_arming_disarming_report" / Flag,
                            "alternate_dialing" / Flag,
                            "dial_tone_delay" / Flag,
                            "report_zone_restore" / Flag,
                        )
                    ),
                    Embedded(
                        Struct(  # EVO section data 3038
                            "access_control_feature" / Flag,
                            "log_request_for_exit" / Flag,
                            "log_door_left_open_restore" / Flag,
                            "log_door_forced_restore" / Flag,
                            "bulglar_alarm_on_forced_door" / Flag,
                            "skip_exit_delay_when_arming_with_access_card" / Flag,
                            "bulglar_alarm_on_door_left_open" / Flag,
                            "who_has_access_during_clock_loss" / Flag,
                        )
                    ),
                )
            ),
            "_not_used1" / Padding(4),
            "source_id" / Default(CommunicationSourceIDEnum, 1),
            "carrier_length" / Bytes(1),
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)

RAMDataParserMap = {
    1: Struct(
        "_weekday" / Int8ub,
        "_system_flags" / BitStruct(  # TODO: Do we need BitsSwapped here?
            "chime_zone_partition" / BitsSwapped(StatusFlags(4)),
            "power_smoke" / Flag,
            "ground_start" / Flag,
            "kiss_off" / Flag,
            "line_ring" / Flag
        ),
        "partition_bell" / BitsSwapped(Bitwise(StatusFlags(8))),
        "partition_fire_alarm" / BitsSwapped(Bitwise(StatusFlags(8))),
        "partition_open_close_kiss_off" / BitsSwapped(Bitwise(StatusFlags(8))),
        "key-switch_triggered" / BitsSwapped(Bitwise(StatusFlags(32))),
        "door_open" / BitsSwapped(Bitwise(StatusFlags(32))),
        "system"
        / Struct(
            "troubles"
            / BitStruct( # time_lost_trouble when actually battery_failure
                "time_lost_trouble" / Flag,
                "zone_fault_trouble" / Flag,
                "zone_low_battery_trouble" / Flag,
                "zone_tamper_trouble" / Flag,
                "module_supervision_trouble" / Flag,  # BusCom
                "module_trouble" / Flag,
                "dialer_trouble" / Flag,
                "system_trouble" / Flag,

                "panel_tamper_trouble" / Flag,
                "_future_use_0" / Flag,
                "rom_error_trouble" / Flag,
                "bell_absent_trouble" / Flag,
                "bell_limit_trouble" / Flag,
                "aux_limit_trouble" / Flag,
                "battery_failure_trouble" / Flag,
                "ac_trouble" / Flag,

                "_future_use_1" / Flag,
                "_future_use_2" / Flag,
                "com_pc_trouble" / Flag,
                "fail_central_4_trouble" / Flag,
                "fail_central_3_trouble" / Flag,
                "fail_central_2_trouble" / Flag,
                "fail_central_1_trouble" / Flag,
                "tlm_trouble" / Flag,

                "module_aux_trouble" / Flag,
                "module_battery_fail" / Flag,
                "module_ac_trouble" / Flag,
                "module_printer_trouble" / Flag,
                "module_fail_to_com_trouble" / Flag,
                "module_tlm_trouble" / Flag,
                "module_rom_error_trouble" / Flag,
                "module_tamper_trouble" / Flag,

                "mdl_com_error" / Flag,
                "bus_overload_trouble" / Flag,
                "bus_global_fail" / Flag,
                "safety_mismatch_trouble" / Flag,
                "_future_use_3" / Flag,
                "_future_use_4" / Flag,
                "missing_module_trouble" / Flag,
                "missing_keypad_trouble" / Flag,
            ),
            "date"
            / Struct(
                "weekday" / Computed(lambda ctx: ctx._._._weekday),
                "time" / DateAdapter(Bytes(7)),
            ),
            "power"
            / Struct(
                "vdc"
                / ExprAdapter(
                    Byte, lambda obj, ctx: round(obj * (20.3 - 1.4) / 255.0 + 1.4, 1), 0
                ),
                "battery"
                / ExprAdapter(Byte, lambda obj, ctx: round(obj * 22.8 / 255.0, 1), 0),
                "dc"
                / ExprAdapter(Byte, lambda obj, ctx: round(obj * 22.8 / 255.0, 1), 0),
            ),
        ),
        "zone_open" / BitsSwapped(Bitwise(StatusFlags(96))),
        "zone_tamper" / BitsSwapped(Bitwise(StatusFlags(96))),
        "zone_low_battery" / BitsSwapped(Bitwise(StatusFlags(96))),
    ),
    2: Struct("zone_status" / ZoneFlags(64)),
    3: Struct(
        "zone_status" / ZoneFlags(32, start_index_from=65),
        "partition_status" / PartitionStatus(Bytes(32)),
    ),
    4: Struct(
        "partition_status" / PartitionStatus(Bytes(16)),
        "system"
        / Struct(
            "panel_status"
            / BitStruct("installer_lock_active" / Flag, "_free" / Padding(7)),
            "event"
            / Struct("_event_pointer" / Int16ub, "_event_pointer_bus" / Int16ub,),
            "_recycle_system" / Array(8, Int8ub),
            "report" / Struct("arm_disarm_delay_timer" / Int8ub,),
        ),
        "_free" / Padding(34),
    ),
    5: Struct(
        "_not_used" / Int8ub,
        "module_trouble" / ModuleTroubles(count=63, start_index_from=1),
    ),
    6: Struct("module_trouble" / ModuleTroubles(count=64, start_index_from=64)),
    7: Struct("module_trouble" / ModuleTroubles(count=64, start_index_from=128)),
    8: Struct(
        "module_trouble" / ModuleTroubles(count=63, start_index_from=192),
        "_not_used" / Int8ub,
    ),
    9: Struct(
        "zone_open" / BitsSwapped(Bitwise(StatusFlags(96, start_index_from=97))),
        "zone_tamper" / BitsSwapped(Bitwise(StatusFlags(96, start_index_from=97))),
        "zone_low_battery" / BitsSwapped(Bitwise(StatusFlags(96, start_index_from=97))),
        "zone_status" / ZoneFlags(28, start_index_from=97),
    ),
    10: Struct("zone_status" / ZoneFlags(64, start_index_from=125)),
    11: Struct(
        "zone_status" / ZoneFlags(4, start_index_from=189), "_not_used" / Bytes(60),
    ),
    16: Struct( # TODO: here should be panel modules
        "module_assigned" / BitsSwapped(Bitwise(StatusFlags(256, start_index_from=1))),
        "module_missing" / BitsSwapped(Bitwise(StatusFlags(256, start_index_from=1))),
    ),
    # 51: Doors [open, state, ...]
    # 56: Disarm delays
    57: Struct("pgm_status" / PGMFlags(16)),
    58: Struct("pgm_status" / PGMFlags(16, start_index_from=17)),
}
# We also need parsers for:
# 17 ram address - EBUS Troubles 24 bytes
# 32 ram address
# 33 ram address
# 34 ram address
# 35 ram address
# 37 ram address
# 38 ram address
# 39 ram address
# 48 ram address
# 49 ram address
# 50 ram address
# 51 ram address - Door Status
# 56 ram address
# 58 ram address


def get_user_definition(settings):
    if (
        settings.system_options.user_code_length_6
        or settings.system_options.user_code_length_flexible
    ):
        code = ExprAdapter(
            Bytes(3),
            lambda obj, path: binascii.hexlify(obj)
            .decode()
            .rstrip("0")
            .replace("a", "0")
            or None,
            lambda obj, path: binascii.unhexlify(obj.replace("0", "a")),
        )
    else:
        code = ExprAdapter(
            Bytes(3),
            lambda obj, path: binascii.hexlify(obj)
            .decode()
            .rstrip("0")
            .replace("a", "0")[:4]
            or None,
            lambda obj, path: binascii.unhexlify((obj + obj[:2]).replace("0", "a")),
        )

    return Struct(
        "code" / code,
        "options"
        / BitsSwapped(
            BitStruct(
                "type" / Enum(BitsInteger(2), FullMaster=0x3, Master=0x2, Regular=0x0),
                "duress" / Flag,
                "bypass" / Flag,
                "arm_only" / Flag,
                "stay_instant_arming" / Flag,
                "force_arming" / Flag,
                "all_subsystems" / Flag,
            )
        ),
        "partitions" / BitsSwapped(Bitwise(StatusFlags(8))),
        "access" / BitStruct("level" / Nibble, "schedule" / Nibble),
        "access_options" / Bytes(1),
        "card_serial_number" / Bytes(3),
    )


DefinitionsParserMap = {
    "zone": BitStruct(
        "definition"
        / Enum(
            Nibble,
            disabled=0x0,
            entry_delay1=0x1,
            entry_delay2=0x2,
            follow=0x3,
            instant=0x4,
            buzzer_24h=0x5,
            burglary_24h=0x6,
            holdup_24h=0x7,
            gas_24h=0x8,
            heat_24h=0x9,
            water_24h=0xA,
            freeze_24h=0xB,
            delayed_fire_24h=0xC,
            standard_fire_24h=0xD,
            stay_delay1=0xE,
            stay_delay2=0xF,
        ),
        "partition" / Nibble,
        "options"
        / ByteSwapped(
            Struct(
                "auto_zone_shutdown_enabled" / Flag,
                "bypass_enabled" / Flag,
                "stay_zone" / Flag,
                "force_zone" / Flag,
                "alarm_type"
                / Enum(
                    BitsInteger(2),
                    steady_alarm=0x0,
                    pulsed_alarm=0x1,
                    silent_alarm=0x2,
                    report_only=0x3,
                ),
                "intellizone" / Flag,
                "delay_before_transmission" / Flag,
            )
        ),
    ),
    "partition": BitsSwapped(  # No need as we get this data during connect in InitializeCommunication system_options
        Bitwise(
            DictArray(
                8,
                1,
                Struct(
                    "_index" / Computed(this._index + 1),
                    "definition"
                    / ExprAdapter(
                        Default(Flag, False),
                        lambda obj, context: "enabled" if obj else "disabled",
                        lambda obj, context: obj == "enabled",
                    ),
                ),
            )
        )
    ),
    "user": get_user_definition,
}

LiveEvent = Struct(
    "fields"
    / RawCopy(
        Struct(
            "po"
            / BitStruct(
                "command" / Const(0xE, Nibble),
                "status"
                / Struct(
                    "reserved" / Flag,
                    "alarm_reporting_pending" / Flag,
                    "Winload_connected" / Flag,
                    "NeWare_connected" / Flag,
                ),
            ),
            "event_source" / Const(0xFF, Int8ub),
            "event_nr" / Int16ub,
            "time" / DateAdapter(Bytes(6)),
            "event" / EventAdapter(Bytes(4)),
            "partition" / Computed(this.event.partition),
            "module_serial" / Bytes(4),
            "label_type" / Bytes(1),
            "label" / Bytes(16),
            "_not_used0" / Bytes(1),
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)

# "Event 1 requested (in compressed format) 1 (12 bytes)
#   Byte 00: [7-3]: Day, [2-0]: Month (MSB)
#   Byte 01: [7]: Month (LSB), [6-0]: Century
#   Byte 02: [7-1]: Year, [0]: Hour (MSB)
#   Byte 03: [7-4]: Hour (LSB), [3-0]: Minutes (MSB)
#   Byte 04: [7-6]: Minutes (LSB), [5-0]: Event Group
#   Byte 05: [7-4]: Partition, [3-0]: Event Number High Nibble
#   Byte 06: Event Number 1
#   Byte 07: Event Number 2
#   Byte 08: Serial Number 1 & 2 (2 nibbles)
#   Byte 09: Serial Number 3 & 4 (2 nibbles)
#   Byte 10: Serial Number 5 & 6 (2 nibbles)
#   Byte 11: Serial Number 7 & 8 (2 nibbles)"

CompressedEvent = Struct(
    "compressed"
    / BitStruct(
        "day" / BitsInteger(5),
        "month" / BitsInteger(4),
        "century" / BitsInteger(7),
        "year" / BitsInteger(7),
        "hour" / BitsInteger(5),
        "minute" / BitsInteger(6),
        "event_group" / BitsInteger(6),
        "partition" / BitsInteger(4),
        "event_1_high_nibble" / BitsInteger(2),
        "event_2_high_nibble" / BitsInteger(2),
    ),
    "minor_1" / Int8ub,
    "minor_2" / Int8ub,
    "module_serial" / Bytes(4),
)

RequestedEvent = Struct(
    "fields"
    / RawCopy(
        Struct(
            "po"
            / BitStruct(
                "command" / Const(0xE, Nibble),
                "status"
                / Struct(
                    "reserved" / Flag,
                    "alarm_reporting_pending" / Flag,
                    "Winload_connected" / Flag,
                    "NeWare_connected" / Flag,
                ),
            ),
            "length" / PacketLength(Int8ub),
            "_not_used0" / Bytes(1),
            "requested_event_nr" / Const(0x00, Int8ub),
            "event_nr" / Int16ub,
            # "data" / Bytes(lambda this: this.length - 7)
            "data" / Array(lambda x: int((x.length - 7) / 12), CompressedEvent),
        )
    ),
    "checksum"
    / Checksum(Bytes(1), lambda data: calculate_checksum(data), this.fields.data),
)

CloseConnection = Struct(
    "fields"
    / RawCopy(
        Struct(
            "po" / Struct("command" / Const(0x70, Int8ub)),
            "length" / PacketLength(Int8ub),
            "message"
            / Default(
                Enum(
                    Int8ub,
                    requested_command_failed=0x00,
                    invalid_user_code=0x01,
                    partition_in_code_lockout=0x02,
                    panel_will_disconnect=0x05,
                    panel_not_connected=0x10,
                    panel_already_connected=0x11,
                    invalid_pc_password=0x12,
                    winload_on_phone_line=0x13,
                    invalid_module_address=0x14,
                    cannot_write_in_ram=0x15,
                    upgrade_request_fail=0x16,
                    record_number_out_of_range=0x17,
                    invalid_record_type=0x19,
                    multibus_not_supported=0x1A,
                    incorrect_number_of_users=0x1B,
                    invalid_label_number=0x1C,
                ),
                0x05,
            ),
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)


class EvoEEPROMAddressAdapter(Subconstruct):
    def deep_update(self, d, u):
        for k, v in u.items():
            if isinstance(v, Mapping):
                d[k] = self.deep_update(d.get(k, {}), v)
            else:
                d[k] = v
        return d

    def _build(self, obj, stream, context, path):
        if "control" in obj and obj["control"].get("ram_access"):  # for ram block
            ram_block = (obj["address"] & 0xF0000) >> 16
            self.deep_update(obj, dict(po=dict(block=ram_block)))
        else:  # for eeprom
            if obj["address"] >> 16 > 0x3:
                raise ValidationError("EEPROM address is out of range")
            eeprom_address_high_bits = (obj["address"] & 0x30000) >> 16
            self.deep_update(
                obj, dict(control=dict(_eeprom_address_bits=eeprom_address_high_bits))
            )
        return self.subcon._build(obj, stream, context, path)

    def _parse(self, stream, context, path):
        obj = self.subcon._parsereport(stream, context, path)
        obj.address += obj.control._eeprom_address_bits << 16
        return obj


ReadEEPROM = Struct(
    "fields"
    / RawCopy(
        EvoEEPROMAddressAdapter(
            Struct(
                "po"
                / BitStruct(
                    "command" / Const(0x5, Nibble), "block" / Default(Nibble, 0),
                ),
                "packet_length" / PacketLength(Int8ub),
                "control"
                / BitStruct(
                    "ram_access" / Default(Flag, False),
                    "alarm_reporting_pending" / Default(Flag, False),
                    "Winload_connected" / Default(Flag, False),
                    "NeWare_connected" / Default(Flag, False),
                    "_not_used" / Default(BitsInteger(2), 0),
                    "_eeprom_address_bits" / Default(BitsInteger(2), 0),
                ),
                "bus_address" / Default(Int8ub, 0x00),  # 00 - Panel, 01-FF - Modules
                "address" / ExprSymmetricAdapter(Int16ub, obj_ & 0xFFFF),
                "length" / Int8ub,
            )
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)

ReadEEPROMResponse = Struct(
    "fields"
    / RawCopy(
        EvoEEPROMAddressAdapter(
            Struct(
                "po"
                / BitStruct(
                    "command" / Const(0x5, Nibble),
                    "status"
                    / Struct(
                        "reserved" / Flag,
                        "alarm_reporting_pending" / Flag,
                        "Winload_connected" / Flag,
                        "NeWare_connected" / Flag,
                    ),
                ),
                "packet_length" / PacketLength(Int8ub),
                "control"
                / BitStruct(
                    "ram_access" / Flag,  # RAM = 0 or EEPROM = 1
                    "_not_used" / Padding(5),
                    "_eeprom_address_bits"
                    / BitsInteger(2),  # EEPROM address bit 17 and 16
                ),
                "bus_address" / Default(Int8ub, 0x00),  # 00 - Panel, 01-FE - Modules
                "address" / ExprSymmetricAdapter(Int16ub, obj_ & 0xFFFF),
                "data" / Bytes(lambda x: x.packet_length - 7),
            )
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)

SetTimeDate = Struct(
    "fields"
    / RawCopy(
        Struct(
            "po" / Struct("command" / Const(0x30, Int8ub)),
            "packet_length" / PacketLength(Int8ub),
            "_not_used0" / Padding(4),
            "century" / Int8ub,
            "year" / Int8ub,
            "month" / Int8ub,
            "day" / Int8ub,
            "hour" / Int8ub,
            "minute" / Int8ub,
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)

SetTimeDateResponse = Struct(
    "fields"
    / RawCopy(
        Struct(
            "po"
            / BitStruct(
                "command" / Const(0x3, Nibble),
                "status"
                / Struct(
                    "reserved" / Flag,
                    "alarm_reporting_pending" / Flag,
                    "Winload_connected" / Flag,
                    "NeWare_connected" / Flag,
                ),
            ),
            "length" / Int8ub,
            "_not_used0" / Padding(4),
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)

_PartitionCommandEnum = Enum(
    Nibble,
    none=0,
    arm=2,
    arm_stay=3,
    arm_instant=4,
    arm_force=5,
    disarm=6,
    beep_keypads=8,
)

PerformPartitionAction = Struct(
    "fields"
    / RawCopy(
        Struct(
            "po" / Struct("command" / Const(0x40, Int8ub)),
            "packet_length" / PacketLength(Int8ub),
            "_not_used0" / Padding(4),
            "partitions"
            / Bitwise(
                DictArray(
                    8,
                    1,
                    Struct(
                        "_index" / Computed(this._index + 1),
                        "command" / Default(_PartitionCommandEnum, "none"),
                    ),
                    pick_key="command",
                )
            ),
            "instant" / Default(Flag, False),
            "_not_used1" / Padding(3),
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)

# Used for partitions and PGMs
PerformActionResponse = Struct(
    "fields"
    / RawCopy(
        Struct(
            "po"
            / BitStruct(
                "command" / Const(0x4, Nibble),
                "status"
                / Struct(
                    "reserved" / Flag,
                    "alarm_reporting_pending" / Flag,
                    "Winload_connected" / Flag,
                    "NeWare_connected" / Flag,
                ),
            ),
            "packet_length" / PacketLength(Int8ub),
            "_not_used0" / Padding(4),
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)

ZoneActionBitOperation = Enum(Int8ub, set=0x08, clear=0x00)

PerformZoneAction = Struct(
    "fields"
    / RawCopy(
        Struct(
            "po" / Struct("command" / Const(0xD0, Int8ub)),
            "packet_length" / PacketLength(Int8ub),
            "flags" / ZoneFlagBitStruct,
            "operation" / ZoneActionBitOperation,
            "_not_used" / Padding(2),
            "zones" / BitsSwapped(Bitwise(EnumerationAdapter(Array(192, Flag)))),
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)

PerformZoneActionResponse = Struct(
    "fields"
    / RawCopy(
        Struct(
            "po"
            / BitStruct(
                "command" / Const(0xD, Nibble),
                "status"
                / Struct(
                    "reserved" / Flag,
                    "alarm_reporting_pending" / Flag,
                    "Winload_connected" / Flag,
                    "NeWare_connected" / Flag,
                ),
            ),
            "packet_length" / PacketLength(Int8ub),
            "flags" / ZoneFlagBitStruct,
            "operation" / ZoneActionBitOperation,
            "_not_used" / Padding(2),
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)

_PGMCommandEnum = Enum(Int8ub, release=0, off=1, on=3, on_override=4, off_override=2)

PerformPGMAction = Struct(
    "fields"
    / RawCopy(
        Struct(
            "po" / Struct("command" / Const(0x40, Int8ub)),
            "packet_length" / PacketLength(Int8ub),
            "unknown0" / Const(0x06, Int8ub),
            "_not_used0" / Padding(3),
            "pgms" / BitsSwapped(Bitwise(EnumerationAdapter(Array(32, Flag)))),
            "_not_used1" / Padding(4),
            "command" / _PGMCommandEnum,
            "_not_used2" / Padding(3),
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)

_PGMBroadcastCommandEnum = Enum(
    Int8ub,
    no_change=0,
    override_off=1,  # deactivate
    override_on=2,  # activate
    release_off=3,  # My brain breaks while I am trying to understand this
    release=4,  # activate and follow events
)

PGMBroadcastCommand = DictArray(
    16,
    1,
    Struct(
        "_index" / Computed(this._index + 1),
        "command" / Default(_PGMBroadcastCommandEnum, "no_change"),
    ),
    pick_key="command",
)

BroadcastRequest = Struct(
    "fields"
    / RawCopy(
        Struct(
            "po"
            / BitStruct(
                "command" / Const(0xA, Nibble),
                "module_type" / Default(Flag, False),
                "sub_command"
                / Enum(
                    BitsInteger(3),
                    general_broadcast=0,
                    lcd_message_off=1,
                    lcd_message_low_prority=2,
                    lcd_message_high_prority=3,
                    pgm_override=4,
                ),
            ),
            "packet_length" / PacketLength(Int8ub),
            "bus_address" / Default(Int8ub, 0x00),  # 00 - Panel, 01-FE - Modules
            "_not_used" / Padding(3),
            "data" / Bytes(16),
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)

BroadcastResponse = Struct(
    "fields"
    / RawCopy(
        EvoEEPROMAddressAdapter(
            Struct(
                "po"
                / BitStruct(
                    "command" / Const(0xA, Nibble),
                    "status"
                    / Struct(
                        "reserved" / Flag,
                        "alarm_reporting_pending" / Flag,
                        "Winload_connected" / Flag,
                        "NeWare_connected" / Flag,
                    ),
                ),
                "packet_length" / PacketLength(Int8ub),
                "bus_address" / Default(Int8ub, 0x00),  # 00 - Panel, 01-FE - Modules
                "control"
                / BitStruct(
                    "ram_access" / Default(Flag, False),
                    "_not_used" / Default(BitsInteger(5), 0),
                    "_eeprom_address_bits" / Default(BitsInteger(2), 0),
                ),
                "address" / ExprSymmetricAdapter(Int16ub, obj_ & 0xFFFF),
            )
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)

ErrorMessage = Struct(
    "fields"
    / RawCopy(
        Struct(
            "po"
            / BitStruct(
                "command" / Const(0x7, Nibble),
                "status"
                / Struct(
                    "reserved" / Flag,
                    "alarm_reporting_pending" / Flag,
                    "Winload_connected" / Flag,
                    "NeWare_connected" / Flag,
                ),
            ),
            "length" / PacketLength(Int8ub),
            "message"
            / Enum(
                Int8ub,
                requested_command_failed=0x00,
                invalid_user_code=0x01,
                partition_in_code_lockout=0x02,
                panel_will_disconnect=0x05,
                panel_not_connected=0x10,
                panel_already_connected=0x11,
                invalid_pc_password=0x12,
                winload_on_phone_line=0x13,
                invalid_module_address=0x14,
                cannot_write_in_ram=0x15,
                upgrade_request_fail=0x16,
                record_number_out_of_range=0x17,
                invalid_record_type=0x19,
                multibus_not_supported=0x1A,
                incorrect_number_of_users=0x1B,
                invalid_label_number=0x1C,
            ),
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)

SendPanicAction = Struct(  # Supported on firmware versions 7.15+
    "fields"
    / RawCopy(
        Struct(
            "po" / Struct("command" / Const(0x40, Int8ub)),
            "packet_length" / PacketLength(Int8ub),
            "unknown0" / Const(0x09, Int8ub),
            "_not_used" / Padding(3),
            "user_id" / Int16ub,
            "panic_type" / Enum(Int8ub, emergency=0, medical=1, fire=2,),  # wild guess
            "partitions" / BitsSwapped(Bitwise(EnumerationAdapter(Array(8, Flag)))),
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)

PerformDoorAction = Struct(
    "fields"
    / RawCopy(
        Struct(
            "po" / Struct("command" / Const(0x40, Int8ub)),
            "packet_length" / PacketLength(Int8ub),
            "unknown0" / Const(0x2, Int8ub),
            "_not_used0" / Padding(3),
            "doors" / BitsSwapped(Bitwise(EnumerationAdapter(Array(32, Flag)))),
            "command" / Enum(Int8ub, lock=1, unlock=2),
            "_not_used1" / Padding(2),
            "unknown1" / Const(0x55, Int8ub),
            "user_id" / Default(Int16ul, 0),
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)),
)



import asyncio
import binascii
import hashlib
import json
import logging
import os
import typing
from collections import namedtuple

from paho.mqtt.client import Client, MQTTMessage

from paradox.config import config as cfg
from paradox.event import Change, Event, EventLevel, Notification
from paradox.lib import ps
from paradox.lib.utils import (JSONByteEncoder, call_soon_in_main_loop,
                               sanitize_key)

from .core import AbstractMQTTInterface
from .helpers import ELEMENT_TOPIC_MAP, get_control_topic_prefix

logger = logging.getLogger("PAI").getChild(__name__)

ParsedMessage = namedtuple("parsed_message", "topics element content")


def mqtt_handle_decorator(
    func: typing.Callable[
        ["BasicMQTTInterface", ParsedMessage],
        typing.Coroutine[None, "BasicMQTTInterface", ParsedMessage],
    ]
):
    async def try_func(*args, **kwargs):
        try:
            return await func(*args, **kwargs)
        except:
            logger.exception("Exception executing MQTT function")

    def wrapper(
        self: "BasicMQTTInterface", client: Client, userdata, message: MQTTMessage
    ):
        try:
            if message.retain:
                logger.warning(
                    "Ignoring command: retained message topic={}, payload={}".format(
                        message.topic, str(message.payload.decode("utf-8"))
                    )
                )
                return

            if self.alarm is None:
                logger.warning(
                    "No alarm. Ignoring command: message topic={}, payload={}".format(
                        message.topic, str(message.payload.decode("utf-8"))
                    )
                )
                return

            logger.info(
                "message topic={}, payload={}".format(
                    message.topic, str(message.payload.decode("utf-8"))
                )
            )

            topic = message.topic.split(cfg.MQTT_BASE_TOPIC)[1]

            topics = topic.split("/")

            if len(topics) < 3:
                logger.error("Invalid topic in mqtt message: {}".format(message.topic))
                return

            content = message.payload.decode("utf-8").strip()

            element = None
            if len(topics) >= 4:
                element = topics[3]

            call_soon_in_main_loop(
                try_func(self, ParsedMessage(topics, element, content))
            )
        except:
            logger.exception("Failed to execute command")

    return wrapper


def _extract_command_user(command):
    aux = command.split(" ")
    if len(aux) > 1:
        return aux[0], aux[1]
    else:
        return aux[0], None


class BasicMQTTInterface(AbstractMQTTInterface):
    def __init__(self, alarm):
        super().__init__(alarm)

        self.partitions = {}
        self.definitions = {}
        self.labels = {}
        self.challenges = None
        self.connected_future = (
            asyncio.Future()
        )  # TODO: do not create it, use some other
        self.labels_future = asyncio.Future()
        self.definitions_future = asyncio.Future()

        ready_future = asyncio.ensure_future(
            asyncio.gather(
                self.connected_future, self.labels_future, self.definitions_future
            )
        )
        ready_future.add_done_callback(lambda x: self._on_ready())

        ps.subscribe(self._handle_panel_labels, "labels_loaded")
        ps.subscribe(self._handle_panel_definitions, "definitions_loaded")
        ps.subscribe(self._handle_panel_change, "changes")
        ps.subscribe(self._handle_panel_event, "events")

    def on_connect(self, mqttc, userdata, flags, result):
        self.subscribe_callback(get_control_topic_prefix("output")+"/#",
            self._mqtt_handle_output_control,
        )
        self.subscribe_callback(
            get_control_topic_prefix("door")+"/#",
            self._mqtt_handle_door_control,
        )
        self.subscribe_callback(
            get_control_topic_prefix("zone")+"/#",
            self._mqtt_handle_zone_control,
        )
        self.subscribe_callback(
            get_control_topic_prefix("partition")+"/#",
            self._mqtt_handle_partition_control,
        )
        self.subscribe_callback(
            "{}/{}/{}".format(cfg.MQTT_BASE_TOPIC, cfg.MQTT_NOTIFICATIONS_TOPIC, "+"),
            self._mqtt_handle_notifications,
        )
        self.subscribe_callback(
            "{}/{}/{}/{}".format(
                cfg.MQTT_BASE_TOPIC, cfg.MQTT_SEND_PANIC_TOPIC, "+", "+"
            ),
            self._mqtt_handle_send_panic,
        )

        if not self.connected_future.done():
            self.connected_future.set_result(True)

    @mqtt_handle_decorator
    async def _mqtt_handle_notifications(self, prep: ParsedMessage):
        topics = prep.topics
        try:
            level = EventLevel.from_name(topics[2].upper())
        except Exception as e:
            logger.error(e)
            return

        ps.sendNotification(
            Notification(sender=self.name, message=prep.content, level=level)
        )

    @mqtt_handle_decorator
    async def _mqtt_handle_zone_control(self, prep: ParsedMessage):
        topics, element, command = prep

        if cfg.MQTT_CHALLENGE_SECRET is not None:
            command, user = self._validate_command_with_challenge(command)
            if command is None:
                return
        else:
            command, user = _extract_command_user(command)

        message = "Zone command: {}={} user: {}".format(element, command, user)
        logger.debug(message)
        self._publish_command_status(message)

        if not await self.alarm.control_zone(element, command):
            message = "Zone command refused: {}={} user: {}".format(
                element, command, user
            )
            logger.warning(message)
        else:
            message = "Zone command accepted: {}={} user: {}".format(
                element, command, user
            )

        self._publish_command_status(message)

    @mqtt_handle_decorator
    async def _mqtt_handle_partition_control(self, prep: ParsedMessage):
        topics, element, command = prep

        if cfg.MQTT_CHALLENGE_SECRET is not None:
            command, user = self._validate_command_with_challenge(command)
            if command is None:
                return
        else:
            command, user = _extract_command_user(command)

        command = cfg.MQTT_COMMAND_ALIAS.get(command, command)

        if command.startswith("code_toggle-"):
            tokens = command.split("-")
            if len(tokens) < 2:
                logger.warning("Invalid token length {}".format(len(tokens)))
                return

            if tokens[1] not in cfg.MQTT_TOGGLE_CODES:
                logger.warning("Invalid toggle code {}".format(tokens[1]))
                return

            if element.lower() == "all":
                command = "arm"

                for k, v in self.partitions.items():
                    # If "all" and a single partition is armed, default is
                    # to disarm
                    for k1, v1 in self.partitions[k].items():
                        if (
                            k1 == "arm" or k1 == "exit_delay" or k1 == "entry_delay"
                        ) and v1:
                            command = "disarm"
                            break

                    if command == "disarm":
                        break

            elif element in self.partitions:
                if (
                    "arm" in self.partitions[element]
                    and self.partitions[element]["arm"]
                ) or (
                    "exit_delay" in self.partitions[element]
                    and self.partitions[element]["exit_delay"]
                ):
                    command = "disarm"
                else:
                    command = "arm"
            else:
                logger.warning("Element {} not found".format(element))
                return

            ps.sendNotification(
                Notification(
                    sender="mqtt",
                    message="Command by {}: {}".format(
                        cfg.MQTT_TOGGLE_CODES[tokens[1]], command
                    ),
                    level=EventLevel.INFO,
                )
            )

        message = "Partition command: {}={} user: {}".format(element, command, user)
        logger.info(message)
        self._publish_command_status(message)

        if not await self.alarm.control_partition(element, command):
            message = "Partition command refused: {}={} user: {}".format(
                element, command, user
            )
            logger.warning(message)
        else:
            message = "Partition command accepted: {}={} user: {}".format(
                element, command, user
            )

        self._publish_command_status(message)

    @mqtt_handle_decorator
    async def _mqtt_handle_output_control(self, prep: ParsedMessage):
        topics, element, command = prep

        if cfg.MQTT_CHALLENGE_SECRET is not None:
            command, user = self._validate_command_with_challenge(command)

            if command is None:
                return
        else:
            command, user = _extract_command_user(command)

        message = "Output command: {}={} user: {}".format(element, command, user)
        logger.debug(message)
        self._publish_command_status(message)

        if not await self.alarm.control_output(element, command):
            message = "Output command refused: {}={} user: {}".format(
                element, command, user
            )
            logger.warning(message)
        else:
            message = "Output command accepted: {}={} user: {}".format(
                element, command, user
            )

        self._publish_command_status(message)

    @mqtt_handle_decorator
    async def _mqtt_handle_send_panic(self, prep: ParsedMessage):
        topics, partition, userid = prep

        panic_type = topics[2]

        if cfg.MQTT_CHALLENGE_SECRET is not None:
            userid, user = self._validate_command_with_challenge(userid)

            if userid is None:
                return
        else:
            userid, user = _extract_command_user(userid)

        message = "Send panic command: partition: {}, userid: {}, user: {}, type: {}".format(
            partition, userid, user, panic_type
        )
        logger.debug(message)
        self._publish_command_status(message)

        if not await self.alarm.send_panic(partition, panic_type, userid):
            message = "Send panic command refused: {}, userid: {}, user: {}, type: {}".format(
                partition, userid, user, panic_type
            )

            logger.warning(message)
        else:
            message = "Send panic command accepted: {}, userid: {}, user: {}, type: {}".format(
                partition, userid, user, panic_type
            )

        self._publish_command_status(message)

    @mqtt_handle_decorator
    async def _mqtt_handle_door_control(self, prep: ParsedMessage):
        topics, element, command = prep

        if cfg.MQTT_CHALLENGE_SECRET is not None:
            command, user = self._validate_command_with_challenge(command)

            if command is None:
                return
        else:
            command, user = _extract_command_user(command)

        message = "Door command: {}={} user=".format(element, command, user)

        logger.debug(message)
        self._publish_command_status(message)

        if not await self.alarm.control_door(element, command):
            message = "Door command refused: {}={} user: {}".format(
                element, command, user
            )
            logger.warning(message)
        else:
            message = "Door command accepted: {}={} user: {}".format(
                element, command, user
            )

        self._publish_command_status(message)

    def _handle_panel_event(self, event: Event):
        """
        Handle Live Event

        :param raw: object with properties (can have byte properties)
        :return:
        """

        if cfg.MQTT_PUBLISH_RAW_EVENTS:
            self.publish(
                "{}/{}/{}".format(
                    cfg.MQTT_BASE_TOPIC, cfg.MQTT_EVENTS_TOPIC, cfg.MQTT_RAW_TOPIC
                ),
                json.dumps(
                    event.props,
                    ensure_ascii=False,
                    cls=JSONByteEncoder,
                    default=str,
                    sort_keys=True,
                ),
                0,
                cfg.MQTT_RETAIN,
            )

    def _handle_panel_labels(self, data: dict):
        self.labels = data

        if not self.labels_future.done():
            self.labels_future.set_result(data)

    def _handle_panel_definitions(self, data: dict):
        self.definitions = data

        if not self.definitions_future.done():
            self.definitions_future.set_result(data)

    def _handle_panel_change(self, change: Change):
        attribute = change.property
        label = change.key
        value = change.new_value
        element_type = change.type

        """Handle Property Change"""
        if label not in self.partitions:
            self.partitions[label] = dict()

        self.partitions[label][attribute] = value
        self._publish(
            f"{cfg.MQTT_BASE_TOPIC}/{cfg.MQTT_STATES_TOPIC}",
            element_type,
            label,
            attribute,
            value,
        )

    def _publish(
        self,
        base: str,
        element_type: str,
        label: str,
        attribute: str,
        value: [str, int, bool],
    ):
        if element_type in ELEMENT_TOPIC_MAP:
            element_topic = ELEMENT_TOPIC_MAP[element_type]
        else:
            element_topic = element_type

        if isinstance(value, dict):
            # This is fragile...
            if "/" in attribute and not attribute.startswith("/"):
                attribute = f"/{attribute}"

            for attr_name, attr_value in value.items():
                label_tp = f"{attribute}/{attr_name}"
                self._publish(base, element_type, label, label_tp, attr_value)
            return

        if cfg.MQTT_USE_NUMERIC_STATES:
            try:
                publish_value = int(value)
            except (TypeError, ValueError):
                logger.debug('Conversion int(%s) failed, use original value', value)
                publish_value = value
        else:
            publish_value = value

        self.publish(
            "{}/{}/{}/{}".format(base, element_topic, sanitize_key(label), attribute),
            "{}".format(publish_value),
            qos=cfg.MQTT_QOS,
            retain=cfg.MQTT_RETAIN,
        )

    def _on_ready(self):
        if cfg.MQTT_PUBLISH_DEFINITIONS:
            for element_type in self.definitions:  # zones, partitions
                labels = self.labels[element_type]
                definitions = self.definitions[element_type]
                for i in definitions:  # numeric index
                    if i not in labels:
                        continue

                    definition = {**(definitions[i]), "id": i}
                    for attribute in definition:  # attribute
                        if element_type == "user" and attribute == "code":
                            continue
                        self._publish(
                            f"{cfg.MQTT_BASE_TOPIC}/{cfg.MQTT_DEFINITION_TOPIC}",
                            element_type,
                            labels[i]["key"],
                            attribute,
                            definition[attribute],
                        )

        if (
            cfg.MQTT_DASH_PUBLISH
            and len(list(self.labels.get("partition", {}).keys())) >= 2
        ):
            self._publish_dash(cfg.MQTT_DASH_TEMPLATE, self.labels.get("partition", {}))

        if cfg.MQTT_CHALLENGE_SECRET:
            logger.info("MQTT Commands authentication enabled")
            self._refresh_challenge()

    def _publish_dash(self, fname, partitions):
        # TODO: move to a separate component
        if len(list(partitions.keys())) < 2:
            return

        if os.path.exists(fname):
            with open(fname, "r") as f:
                data = f.read()

                for k in partitions.keys():
                    data = data.replace(f"__PARTITION{k}__", partitions[k]["label"])

                self.publish(cfg.MQTT_DASH_TOPIC, data, 2, True)
                logger.info(
                    "MQTT Dash panel published to {}".format(cfg.MQTT_DASH_TOPIC)
                )
        else:
            logger.warning("MQTT DASH Template not found: {}".format(fname))

    def _refresh_challenge(self):
        self.challenge = binascii.hexlify(os.urandom(16))
        self.publish(
            f"{cfg.MQTT_BASE_TOPIC}/{cfg.MQTT_STATES_TOPIC}/{cfg.MQTT_CHALLENGE_TOPIC}",
            self.challenge,
            2,
            True,
        )

    def _validate_command_with_challenge(self, command):
        aux = command.strip().split(" ")

        challenge = self.challenge
        self._refresh_challenge()

        if challenge is None:
            logger.warning("No challenge set")
            return None, None

        if len(aux) != 2 and len(aux) != 3:
            logger.warning("Invalid command format. Authentication code required")
            return None, None

        user = None
        response = None

        if isinstance(cfg.MQTT_CHALLENGE_SECRET, dict):
            if aux[1] in cfg.MQTT_CHALLENGE_SECRET and len(aux) == 3:
                secret = cfg.MQTT_CHALLENGE_SECRET[aux[1]]
                user = aux[1]
                response = aux[2]
        elif isinstance(cfg.MQTT_CHALLENGE_SECRET, str):
            secret = cfg.MQTT_CHALLENGE_SECRET
            response = aux[1]
        else:
            logger.error(
                f"Authentication failed. Invalid setting MQTT_CHALLENGE_SECRET of type {type(cfg.MQTT_CHALLENGE_SECRET)}"
            )
            return None, None

        h = hashlib.new("SHA1")
        i = cfg.MQTT_CHALLENGE_ROUNDS
        text = f"{challenge}{secret}".encode("utf-8")

        while i > 0:
            h.update(text)
            i -= 1

        if response == h.hexdigest():
            message = f"Authentication success. user: {user}"
            ret = aux[0]  # Pass the command
        else:
            message = f"Authentication failed. user: {user}"
            ret = None
            user = None  # Clear user

        self._publish_command_status(message)
        logger.info(message)
        return ret, user



from paradox.config import config as cfg
from paradox.interfaces.mqtt.entities.device import Device
from paradox.interfaces.mqtt.helpers import (
    get_control_topic_prefix,
    get_state_topic_prefix,
)


def to_label(txt):
    return txt.replace("_", " ").title()


class AbstractEntity:
    def __init__(self, device: Device, availability_topic):
        self.availability_topic = availability_topic
        self.device = device

        self.pai_entity_type: str = None
        self.hass_entity_type: str = None

        self.key: str = None
        self.property: str = None
        self.label: str = None

    @property
    def entity_id(self):
        return f"{self.pai_entity_type}_{self.key.lower()}_{self.property}"

    @property
    def entity_name(self):
        label = self.label or to_label(self.key)

        return f"{self.pai_entity_type.title()} {label} {to_label(self.property)}"

    @property
    def configuration_topic(self):
        return "/".join(
            [
                cfg.MQTT_HOMEASSISTANT_DISCOVERY_PREFIX,
                self.hass_entity_type,
                self.device.serial_number,
                self.entity_id,
                "config",
            ]
        )

    def serialize(self):
        prefix = cfg.MQTT_HOMEASSISTANT_ENTITY_PREFIX.format(
            {
                "serial_number",
                self.device.serial_number,
                "model",
                self.device.model,
            }
        )
        return dict(
            availability_topic=self.availability_topic,
            device=self.device,
            name=prefix + f"{self.entity_name}",
            unique_id=f"paradox_{self.device.serial_number}_{self.entity_id}",
            state_topic=self.state_topic,
        )

    @property
    def command_topic(self):
        prefix = get_control_topic_prefix(self.pai_entity_type)
        return f"{prefix}/{self.key}"

    @property
    def state_topic(self):
        prefix = get_state_topic_prefix(self.pai_entity_type)
        return f"{prefix}/{self.key}/{self.property}"


class AbstractControllableEntity(AbstractEntity):
    def serialize(self):
        config = super().serialize()
        config.update(dict(command_topic=self.command_topic))
        return config



import asyncio
from binascii import hexlify
from datetime import datetime
import logging
import time
from typing import Callable, Iterable, Optional

from construct import Container
import pytz

from paradox.config import config as cfg, get_limits_for_type
from paradox.connections.connection import Connection
from paradox.data.enums import RunState
from paradox.data.memory_storage import MemoryStorage as Storage
from paradox.data.model import DetectedPanel
from paradox.event import Change, ChangeEvent, Event, LiveEvent
from paradox.exceptions import (
    AuthenticationFailed,
    CodeLockout,
    PanelNotDetected,
    StatusRequestException,
    async_loop_unhandled_exception_handler,
)
from paradox.hardware import Panel, create_panel
from paradox.lib import ps
from paradox.lib.async_message_manager import ErrorMessageHandler, EventMessageHandler
from paradox.lib.handlers import PersistentHandler
from paradox.lib.utils import deep_merge
from paradox.parsers.status import convert_raw_status

logger = logging.getLogger("PAI").getChild(__name__)


class Paradox:
    def __init__(self, retries=3):
        self.panel: Panel = None
        self._connection: Connection = None
        self.retries = retries
        self.work_loop = asyncio.get_event_loop()
        self.work_loop.set_exception_handler(async_loop_unhandled_exception_handler)

        self.storage = Storage()

        self._run_state = RunState.STOP
        self.request_lock = asyncio.Lock()
        self.busy = asyncio.Lock()
        self.loop_wait_event = asyncio.Event()

        ps.subscribe(self._on_labels_load, "labels_loaded")
        ps.subscribe(self._on_definitions_load, "definitons_loaded")
        ps.subscribe(self._on_status_update, "status_update")
        ps.subscribe(self._on_event, "events")
        ps.subscribe(self._on_property_change, "changes")

    @property
    def run_state(self) -> RunState:
        return self._run_state

    @run_state.setter
    def run_state(self, value: RunState):
        self._run_state = value
        ps.sendMessage("run-state", state=value)

    @property
    def connection(self):
        if not self._connection:
            # Load a connection to the alarm
            if cfg.CONNECTION_TYPE == "Serial":
                logger.info("Using Serial Connection")

                from paradox.connections.serial_connection import SerialCommunication

                self._connection = SerialCommunication(
                    port=cfg.SERIAL_PORT,
                    baud=cfg.SERIAL_BAUD,
                )
            elif cfg.CONNECTION_TYPE == "IP":
                logger.info("Using IP Connection")

                if cfg.IP_CONNECTION_BARE:
                    from paradox.connections.ip.connection import BareIPConnection

                    self._connection = BareIPConnection(
                        host=cfg.IP_CONNECTION_HOST, port=cfg.IP_CONNECTION_PORT
                    )
                elif cfg.IP_CONNECTION_SITEID and cfg.IP_CONNECTION_EMAIL:
                    from paradox.connections.ip.connection import StunIPConnection

                    self._connection = StunIPConnection(
                        site_id=cfg.IP_CONNECTION_SITEID,
                        email=cfg.IP_CONNECTION_EMAIL,
                        panel_serial=cfg.IP_CONNECTION_PANEL_SERIAL,
                        password=cfg.IP_CONNECTION_PASSWORD,
                    )
                else:
                    from paradox.connections.ip.connection import LocalIPConnection

                    self._connection = LocalIPConnection(
                        host=cfg.IP_CONNECTION_HOST,
                        port=cfg.IP_CONNECTION_PORT,
                        password=cfg.IP_CONNECTION_PASSWORD,
                    )
            else:
                raise AssertionError(f"Invalid connection type: {cfg.CONNECTION_TYPE}")

            self._register_connection_handlers()

        return self._connection

    def _register_connection_handlers(self):
        self.connection.register_raw_handler(
            PersistentHandler(self.on_connection_message)
        )

        self.connection.register_handler(EventMessageHandler(self.handle_event_message))
        self.connection.register_handler(ErrorMessageHandler(self.handle_error_message))

    async def connect(self) -> bool:
        if self._connection:
            await self.disconnect()  # socket needs to be also closed
        self.panel = None

        self.run_state = RunState.INIT
        logger.info("Connecting to interface")
        if not await self.connection.connect():
            self.run_state = RunState.ERROR
            logger.error("Failed to connect to interface")
            return False

        logger.info("Connecting to Panel")

        if not self.panel:
            self.panel = create_panel(self)
            self.connection.variable_message_length(self.panel.variable_message_length)

        try:
            initiate_reply = await self.send_wait(
                self.panel.get_message("InitiateCommunication"),
                None,
                reply_expected=0x7,
            )

            if initiate_reply:
                model = initiate_reply.fields.value.label.strip(b"\0 ").decode(
                    cfg.LABEL_ENCODING
                )
                firmware_version = "{}.{} build {}".format(
                    initiate_reply.fields.value.application.version,
                    initiate_reply.fields.value.application.revision,
                    initiate_reply.fields.value.application.build,
                )
                serial_number = hexlify(
                    initiate_reply.fields.value.serial_number
                ).decode()

                logger.info(f"Panel Identified {model} version {firmware_version}")
            else:
                raise ConnectionError("Panel did not replied to InitiateCommunication")

            logger.info("Initiating panel connection")
            reply = await self.send_wait(
                self.panel.get_message("StartCommunication"),
                args=dict(source_id=0x2),
                reply_expected=0x0,
            )

            if reply is None:
                raise ConnectionError("Panel did not replied to StartCommunication")

            if reply.fields.value.product_id is not None:
                self.panel = create_panel(
                    self, reply
                )  # Now we know what panel it is. Let's
                # recreate panel object.
                ps.sendMessage(
                    "panel_detected",
                    panel=DetectedPanel(
                        product_id=reply.fields.value.product_id,
                        model=model,
                        firmware_version=firmware_version,
                        serial_number=serial_number,
                    ),
                )
            else:
                raise PanelNotDetected("Failed to detect panel")

            result = await self.panel.initialize_communication(cfg.PASSWORD)
            if not result:
                raise ConnectionError("Failed to initialize communication")

            self.run_state = RunState.CONNECTED
            logger.info("Connection OK")
            return True
        except asyncio.TimeoutError:
            logger.error(
                "Timeout while connecting to panel. Is an other connection active?"
            )
        except ConnectionError as e:
            logger.error("Failed to connect: %s" % str(e))

        self.run_state = RunState.ERROR

        return False

    async def full_connect(self) -> bool:
        try:
            if not await self.connect():
                return False

            logger.info("Loading data from panel memory")
            await self.panel.load_memory()

            logger.info("Running")
            self.run_state = RunState.RUN
            self.request_status_refresh()  # Trigger status update

            ps.sendMessage("connected")
            return True
        except asyncio.TimeoutError:
            logger.error(
                "Timeout while connecting to panel. Is an other connection active?"
            )
        except ConnectionError as e:
            logger.error("Failed to connect: %s" % str(e))

        self.run_state = RunState.ERROR

        return False

    async def dump_memory(self, file, memory_type):
        if hasattr(self.panel, "dump_memory") and callable(self.panel.dump_memory):
            logger.warning("Requested memory dump. Dumping...")

            await self.panel.dump_memory(file, memory_type)
            logger.warning("Memory dump completed. Exiting pai.")
            raise SystemExit()
        else:
            logger.warning(
                "Requested memory dump, but current panel type does not support it yet."
            )

    async def sync_time(self):
        now = datetime.now().astimezone()
        if cfg.SYNC_TIME_TIMEZONE:
            try:
                tzinfo = pytz.timezone(cfg.SYNC_TIME_TIMEZONE)
                now = now.astimezone(tzinfo)
            except pytz.exceptions.UnknownTimeZoneError:
                logger.debug(
                    f"Panel Timezone Unknown ('{cfg.SYNC_TIME_TIMEZONE}'). Skipping sync"
                )
                return

        if not self._is_time_sync_required(now.replace(tzinfo=None)):
            return

        args = dict(
            century=int(now.year / 100),
            year=int(now.year % 100),
            month=now.month,
            day=now.day,
            hour=now.hour,
            minute=now.minute,
        )

        logger.debug("Synchronizing panel time")
        reply = await self.send_wait(
            self.panel.get_message("SetTimeDate"), args, reply_expected=0x3, timeout=10
        )
        if reply is None:
            logger.warning("Could not set panel time")
        else:
            logger.info("Panel time synchronized")

    async def loop(self):
        logger.debug("Loop start")

        replies_missing = 0

        while self.run_state not in (RunState.STOP, RunState.ERROR):
            tstart = time.time()
            if self.run_state == RunState.RUN:
                try:
                    await self.busy.acquire()
                    result = await asyncio.gather(*self.panel.get_status_requests())
                    merged = deep_merge(*result, extend_lists=True, initializer={})
                    self.work_loop.call_soon(self._process_status, merged)
                    replies_missing = max(0, replies_missing - 1)
                except ConnectionError:
                    raise
                except (StatusRequestException, asyncio.TimeoutError):
                    replies_missing += 1
                    if replies_missing > 3:
                        logger.error("Lost communication with panel")
                        await self.disconnect()
                        return
                except Exception:
                    logger.exception("Loop")
                finally:
                    self.busy.release()

                if replies_missing > 0:
                    logger.debug(f"Loop: Replies missing: {replies_missing}")

            # cfg.Listen for events

            max_wait_time = max((tstart + cfg.KEEP_ALIVE_INTERVAL) - time.time(), 0)
            try:
                await asyncio.wait_for(self.loop_wait_event.wait(), max_wait_time)
            except asyncio.TimeoutError:
                # It is fine to timeout to go to the next loop
                pass
            finally:
                self.loop_wait_event.clear()

    @staticmethod
    def _process_status(raw_status: Container) -> None:
        status = convert_raw_status(raw_status)

        for limit_key in cfg.LIMITS.keys():
            if limit_key not in status:
                continue

            limit_arr = get_limits_for_type(limit_key)
            if limit_arr is not None:
                status[limit_key].filter(limit_arr)

        if cfg.LOGGING_DUMP_STATUS:
            logger.debug("properties: %s", status)

        ps.sendMessage("status_update", status=status)

    def request_status_refresh(self):
        self.loop_wait_event.set()

    def on_connection_message(self, message: bytes):
        if not self.panel:
            return

        try:
            recv_message = self.panel.parse_message(message, direction="frompanel")

            if cfg.LOGGING_DUMP_MESSAGES:
                logger.debug("Message received: %s", recv_message)

            # No message
            if recv_message is None:
                logger.debug(
                    "Unknown message: %s" % (" ".join(f"{c:02x} " for c in message))
                )
                return

            self.connection.schedule_message_handling(
                recv_message
            )  # schedule handling in the loop
        except Exception:
            logger.exception("Error parsing message")

    async def send_wait(
        self,
        message_type=None,
        args=None,
        message=None,
        retries=5,
        timeout=cfg.IO_TIMEOUT,
        reply_expected=None,
    ) -> Optional[Container]:
        # Connection closed
        if not self.connection.connected:
            raise ConnectionError("Not connected")

        if message is None and message_type is not None:
            message = message_type.build(dict(fields=dict(value=args)))

        attempt = 1

        while attempt <= retries:
            if attempt >= 2:  # second and further attempts
                logger.debug("Request retry (attempt %d/%d)", attempt, retries)

            async with self.request_lock:
                t1 = time.time()
                result = "unknown"
                try:
                    if message is not None:
                        self.connection.write(message)

                    if reply_expected is not None:
                        if isinstance(reply_expected, Callable):
                            reply = await self.connection.wait_for_message(
                                reply_expected, timeout=timeout * 2
                            )
                        elif isinstance(reply_expected, Iterable):
                            reply = await self.connection.wait_for_message(
                                lambda m: any(
                                    m.fields.value.po.command == expected
                                    for expected in reply_expected
                                ),
                                timeout=timeout * 2,
                            )
                        else:
                            reply = await self.connection.wait_for_message(
                                lambda m: m.fields.value.po.command == reply_expected,
                                timeout=timeout * 2,
                            )

                        result = "ok"
                        return reply
                except asyncio.TimeoutError:
                    result = "timeout"
                    if attempt == retries:
                        raise
                except ConnectionError:
                    result = "connection error"
                    raise
                except Exception:
                    result = "exception"
                    logger.exception("Unexpected exception during send_wait")
                    raise
                finally:
                    logger.debug("send/receive %s in %.4f s", result, time.time() - t1)

            attempt += 1

        return None  # Probably it needs to throw an exception instead of returning None

    async def control_zone(self, zone: str, command: str) -> bool:
        command = command.lower()
        logger.debug(f"Control Zone: {zone} - {command}")

        zones_selected = self.storage.get_container("zone").select(zone)

        # Not Found
        if len(zones_selected) == 0:
            logger.error("No zones selected")
            return False

        # Apply state changes
        accepted = False
        try:
            accepted = await self.panel.control_zones(zones_selected, command)
        except NotImplementedError:
            logger.error("control_zone is not implemented for this alarm type")
        except asyncio.CancelledError:
            logger.error("control_zone canceled")
        except asyncio.TimeoutError:
            logger.error("control_zone timeout")

        # Refresh status
        self.request_status_refresh()  # Trigger status update

        return accepted

    async def control_partition(self, partition: str, command: str) -> bool:
        command = command.lower()
        logger.debug(f"Control Partition: {partition} - {command}")

        partitions_selected = self.storage.get_container("partition").select(partition)

        # Not Found
        if len(partitions_selected) == 0:
            logger.error("No partitions selected")
            return False

        # Apply state changes
        accepted = False
        try:
            accepted = await self.panel.control_partitions(partitions_selected, command)
        except NotImplementedError:
            logger.error("control_partition is not implemented for this alarm type")
        except asyncio.CancelledError:
            logger.error("control_partition canceled")
        except asyncio.TimeoutError:
            logger.error("control_partition timeout")

        # Refresh status
        self.request_status_refresh()  # Trigger status update

        return accepted

    async def control_output(self, output, command) -> bool:
        command = command.lower()
        logger.debug(f"Control Output: {output} - {command}")

        outputs_selected = self.storage.get_container("pgm").select(output)

        # Not Found
        if len(outputs_selected) == 0:
            logger.error("No outputs selected")
            return False

        # Apply state changes
        accepted = False
        try:
            accepted = await self.panel.control_outputs(outputs_selected, command)
        except NotImplementedError:
            logger.error("control_output is not implemented for this alarm type")
        except asyncio.CancelledError:
            logger.error("control_output canceled")
        except asyncio.TimeoutError:
            logger.error("control_output timeout")
        # Apply state changes

        # Refresh status
        self.request_status_refresh()  # Trigger status update

        return accepted

    async def send_panic(self, partition_id, panic_type, user_id) -> bool:
        logger.debug(
            "Send panic: {}, user: {}, type: {}".format(
                partition_id, user_id, panic_type
            )
        )

        partition = self.storage.get_container_object("partition", partition_id)
        user = self.storage.get_container_object("user", user_id)

        if partition is None or user is None:
            logger.error("Send panic: user or partition is not found")

        try:
            return await self.panel.send_panic(
                [partition["id"]], panic_type, user["id"]
            )
        except NotImplementedError:
            logger.error("send_panic is not implemented for this alarm type")
        except asyncio.CancelledError:
            logger.error("send_panic canceled")
        except asyncio.TimeoutError:
            logger.error("send_panic timeout")

    async def control_door(self, door, command) -> bool:
        command = command.lower()
        logger.debug(f"Control Door: {door} - {command}")

        doors_selected = self.storage.get_container("door").select(door)

        # Not Found
        if len(doors_selected) == 0:
            logger.error("No doors selected")
            return False

        # Apply state changes
        accepted = False
        try:
            accepted = await self.panel.control_doors(doors_selected, command)
        except NotImplementedError:
            logger.error("control_door is not implemented for this alarm type")
        except asyncio.CancelledError:
            logger.error("control_door canceled")
        except asyncio.TimeoutError:
            logger.error("control_door timeout")
        # Apply state changes

        # Refresh status
        self.request_status_refresh()  # Trigger status update

        return accepted

    def get_label(self, label_type: str, label_id) -> Optional[str]:
        el = self.storage.get_container_object(label_type, label_id)
        if el:
            return el.get("label")

    def handle_event_message(self, message: Container = None):
        """Process cfg.Live Event Message and dispatch it to the interface module"""
        try:
            try:
                evt = LiveEvent(
                    event=message,
                    event_map=self.panel.event_map,
                    label_provider=self.get_label,
                )
            except AssertionError:
                logger.debug("Error creating event")
                return

            element = self.storage.get_container_object(evt.type, evt.id)

            # Temporary to catch labels/properties in wrong places
            # TODO: REMOVE
            if message is not None:
                if not evt.id:
                    logger.debug(
                        "Missing element ID in {}/{}, m/m: {}/{}, message: {}".format(
                            evt.type,
                            evt.label or "?",
                            evt.major,
                            evt.minor,
                            evt.message,
                        )
                    )
                else:
                    if not element:
                        logger.warning(
                            "Missing element with ID {} in {}/{}".format(
                                evt.id, evt.type, evt.label
                            )
                        )
                    else:
                        for k in evt.change:
                            if k not in element:
                                logger.warning(
                                    "Missing property {} in {}/{}".format(
                                        k, evt.type, evt.label
                                    )
                                )
                        if evt.label != element.get("label"):
                            logger.warning(
                                "Labels differ {} != {} in {}/{}".format(
                                    element.get("label"), evt.label, evt.type, evt.label
                                )
                            )
            # Temporary end

            # The event has changes. Update the state
            if len(evt.change) > 0 and element:
                self.storage.update_container_object(evt.type, evt.id, evt.change)

            ps.sendEvent(evt)

        except Exception:
            logger.exception("Handle live event")

    def handle_error_message(self, message):
        """Handle ErrorMessage"""
        error_enum = message.fields.value.message

        if error_enum == "panel_not_connected":
            asyncio.get_event_loop().create_task(self.disconnect())
        else:
            message = self.panel.get_error_message(error_enum)
            logger.error(f"Got ERROR Message: {message}")
            if message == "Invalid PC Password":
                raise AuthenticationFailed()
            elif "code lockout" in message:
                raise CodeLockout()

    async def disconnect(self):
        logger.info("Disconnecting from the Alarm Panel")
        self.run_state = RunState.STOP

        self._clean_session()
        if self.connection.connected:
            await self.connection.close()
            logger.info("Disconnected from the Alarm Panel")

    async def pause(self):
        logger.info("Pausing PAI")
        if self.run_state == RunState.RUN:
            logger.info("Pausing from the Alarm Panel")
            self.run_state = RunState.PAUSE
            self.connection.handler_registry.set_ignore_if_no_handlers(True)
            # EVO IP150 IP Interface does not work if we send this
            # await self.send_wait(self.panel.get_message('CloseConnection'), None)

    async def resume(self):
        logger.info("Resuming PAI")
        if self.run_state == RunState.PAUSE:
            self.connection.handler_registry.set_ignore_if_no_handlers(False)
            await self.full_connect()

    def _clean_session(self):
        logger.info("Clean Session")
        if self.connection.connected:
            if not self.panel:
                logger.info("No panel, creating generic one")
                panel = create_panel(self)
            else:
                panel = self.panel

            logger.info("Cleaning previous session. Closing connection")
            # Write directly as this can be called from other contexts

            self.connection.write(panel.get_message("CloseConnection").build(dict()))

    def _on_labels_load(self, data):
        for k, d in data.items():
            self.storage.get_container(k).deep_merge(d)

    def _on_definitions_load(self, data):
        for k, d in data.items():
            self.storage.get_container(k).deep_merge(d)

    def _on_status_update(self, status):
        """
        Calls update_properties
        :param status:
        :return:
        """
        if "troubles" in status:
            self._process_trouble_statuses(status["troubles"])

        for element_type, element_items in status.items():
            if element_type in ["troubles"]:  # troubles was already parsed
                continue
            for element_item_key, element_item_status in element_items.items():
                if isinstance(
                    element_item_status,
                    (
                        dict,
                        list,
                    ),
                ):
                    self.storage.update_container_object(
                        element_type, element_item_key, element_item_status
                    )
                else:
                    logger.debug(
                        "%s/%s:%s ignored",
                        element_type,
                        element_item_key,
                        element_item_status,
                    )

        self._update_partition_states()

        if cfg.SYNC_TIME:
            self.work_loop.create_task(self.sync_time())

    def _process_trouble_statuses(self, trouble_statuses):
        global_trouble = False
        for t_key, t_status in trouble_statuses.items():
            if not isinstance(t_status, bool):
                logger.error("Trouble %s has not boolean state: %s", t_key, t_status)
                continue

            self.storage.update_container_object(
                "system", "troubles", {t_key: t_status}
            )

            global_trouble = global_trouble or t_status

        self.storage.update_container_object(
            "system", "troubles", {"trouble": global_trouble}
        )

    def _is_time_sync_required(self, now) -> bool:
        assert now.tzinfo is None
        try:
            drift = (
                now - self.storage.get_container("system")["date"]["time"]
            ).total_seconds()

            if abs(drift) > cfg.SYNC_TIME_MIN_DRIFT:
                logger.info(f"Time drifted more than allowed: {drift} seconds")
                return True
            else:
                logger.debug(f"Time drifted within allowed range: {drift} seconds")

        except KeyError:
            pass

        return False

    def _update_partition_states(self):
        """
        current_state is fully HomeAssistant compatible. Check HASS manual before making any changes.
        """
        for _, properties in self.storage.get_container("partition").items():
            change = {}
            if any(
                [
                    properties.get("fire_alarm"),
                    properties.get("audible_alarm"),
                    properties.get("silent_alarm"),
                    properties.get("panic_alarm"),
                ]
            ):
                change["current_state"] = "triggered"
            elif properties.get("arm"):
                if properties.get("arm_stay"):
                    change["current_state"] = "armed_home"
                elif properties.get("arm_sleep") or properties.get("arm_no_entry"):
                    change["current_state"] = "armed_night"
                elif properties.get("arm_away"):
                    change["current_state"] = "armed_away"
                else:
                    change["current_state"] = "armed_away"

                change["target_state"] = change["current_state"]
            else:
                change["target_state"] = change["current_state"] = "disarmed"

            if properties.get("exit_delay"):  # Redefine if pending
                change["current_state"] = "arming"

            self.storage.update_container_object("partition", properties["key"], change)

    def _on_event(self, event: Event):
        if cfg.LOGGING_DUMP_EVENTS:
            logger.debug(f"LiveEvent: {event}")

        event.call_hook(storage=self.storage, alarm=self)

        if isinstance(event, LiveEvent):
            self._update_partition_states()

    def _on_property_change(self, change: Change):
        if change.initial:
            return

        try:
            event = ChangeEvent(
                change_object=change,
                property_map=self.panel.property_map,
                label_provider=self.get_label,
            )
            if cfg.LOGGING_DUMP_EVENTS:
                logger.debug(f"ChangeEvent: {event}")
            ps.sendEvent(event)
        except AssertionError:
            logger.debug("Could not create event from change")



# -*- coding: utf-8 -*-

import asyncio
import binascii
import inspect
import logging
import typing

from construct import ChecksumError, Construct, Container

from paradox.config import config as cfg
from paradox.exceptions import AuthenticationFailed, StatusRequestException
from . import parsers
from .event import event_map
from .property import property_map
from ..panel import Panel as PanelBase

logger = logging.getLogger("PAI").getChild(__name__)

PARTITION_ACTIONS = dict(
    arm=0x04,
    disarm=0x05,
    arm_stay=0x01,
    arm_force=0x02,
    arm_sleep=0x03,
    arm_stay_stayd=0x06,
    arm_sleep_stay=0x07,
    disarm_all=0x08,
)
ZONE_ACTIONS = dict(bypass=0x10, clear_bypass=0x10)
PGM_ACTIONS = dict(on_override=0x30, off_override=0x31, on=0x32, off=0x33, pulse=0)


class Panel(PanelBase):

    supports_pgm_status: bool
    event_map = event_map
    property_map = property_map
    max_eeprom_response_data_length = 32

    mem_map = {
        "status_base1": 0x8000,
        "status_base2": 0x1FE0,
        "definitions": {
            "zone": {"addresses": [range(0x730, 0x7A0, 0x03)]},
            "pgm": {"addresses": [range(0x7A0, 0x800, 0x06)]}},
        "labels": {
            "zone": {"label_offset": 0, "addresses": [range(0x010, 0x210, 0x10)]},
            "pgm": {
                "label_offset": 0,
                "addresses": [range(0x210, 0x310, 0x10)],
                "template": {"on": False, "pulse": False},
            },
            "partition": {"label_offset": 0, "addresses": [range(0x310, 0x330, 0x10)]},
            "user": {"label_offset": 0, "addresses": [range(0x330, 0x530, 0x10)]},
            "module": {"label_offset": 0, "addresses": [range(0x530, 0x620, 0x10)]},
            "repeater": {"label_offset": 0, "addresses": [range(0x620, 0x640, 0x10)]},
            "keypad": {"label_offset": 0, "addresses": [range(0x640, 0x6C0, 0x10)]},
            "site": {"label_offset": 0, "addresses": [range(0x6C0, 0x6D0, 0x10)]},
            "siren": {"label_offset": 0, "addresses": [range(0x6D0, 0x700, 0x10)]},
        },
    }

    def __init__(
        self, core, start_communication_response, variable_message_length=True
    ):
        super(Panel, self).__init__(core, variable_message_length)

        self.settings = start_communication_response.fields.value

        self.supports_pgm_status = self.settings.firmware.version >= 6

    @property
    def status_request_addresses(self):
        addresses = list(parsers.RAMDataParserMap.keys())

        if not self.supports_pgm_status:
            addresses.remove(7)

        return addresses

    async def dump_memory(self, file, memory_type):
        """
        Dumps EEPROM and RAM memory to files
        :return:
        """
        if memory_type == "ram":
            await self.dump_memory_to_file(file, range(0, 9), ram=True)
        elif memory_type == "eeprom":
            await self.dump_memory_to_file(file, range(0, 0x0FFF, 32))
        else:
            raise AttributeError(f"Unknown memory type: {memory_type}")

    async def dump_memory_to_file(self, file, range_, ram=False):
        mem_type = "RAM" if ram else "EEPROM"
        logger.info("Dump " + mem_type)

        for address in range_:
            if ram:
                args = dict(address=address + self.mem_map["status_base1"])
            else:
                args = dict(address=address)

            logger.info("Dumping %s: address %x" % (mem_type, address))

            reply = await self.core.send_wait(
                parsers.ReadEEPROM,
                args,
                reply_expected=lambda m: m.fields.value.po.command == 0x5
                and m.fields.value.address == address,
            )

            if reply is None:
                logger.error("Could not read %s: address %x" % (mem_type, address))
                return

            data = reply.fields.value.data

            file.write(data)

    def get_message(self, name: str) -> Construct:
        try:
            clsmembers = dict(inspect.getmembers(parsers))
            if name in clsmembers:
                return clsmembers[name]
        except ResourceWarning:
            pass

        return super(Panel, self).get_message(name)

    def parse_message(
        self, message: bytes, direction="topanel"
    ) -> typing.Optional[Container]:
        try:
            if message is None or len(message) == 0:
                return None

            parent_parsed = super(Panel, self).parse_message(message, direction)
            if parent_parsed:
                return parent_parsed

            if direction == "topanel":
                if message[0] == 0x70 and message[-5] != 0:
                    return parsers.CloseConnection.parse(message)
                elif message[0] == 0x00:
                    return parsers.InitializeCommunication.parse(message)
                elif message[0] == 0x30:
                    return parsers.SetTimeDate.parse(message)
                elif message[0] == 0x40:
                    return parsers.PerformAction.parse(message)
                elif message[0] == 0x50 and message[2] < 0x80:
                    return parsers.ReadEEPROM.parse(message)

            else:
                if message[0] == 0x10:
                    return parsers.InitializeCommunicationResponse.parse(message)
                elif message[0] >> 4 == 0x7 and message[-5] == 0:
                    return parsers.ErrorMessage.parse(message)
                elif message[0] >> 4 == 0x3:
                    return parsers.SetTimeDateResponse.parse(message)
                elif message[0] >> 4 == 0x4:
                    return parsers.PerformActionResponse.parse(message)
                elif message[0] >> 4 == 0x5 and message[2] == 0x80:
                    return parsers.ReadStatusResponse.parse(message)
                elif message[0] >> 4 == 0x5 and message[2] < 0x80:
                    return parsers.ReadEEPROMResponse.parse(message)

                #        elif message[0] == 0x60 and message[2] < 0x80:
                #            return WriteEEPROM.parse(message)
                #        elif message[0] >> 4 == 0x6 and message[2] < 0x80:
                #            return WriteEEPROMResponse.parse(message)
                elif message[0] >> 4 == 0xE:
                    return parsers.LiveEvent.parse(message)

        except ChecksumError as e:
            logger.error(
                "ChecksumError %s, message: %s" % (str(e), binascii.hexlify(message))
            )
        except:
            logger.exception(
                "Exception parsing message: %s" % (binascii.hexlify(message))
            )
        return None

    async def initialize_communication(self, password):
        encoded_password = self.encode_password(password)

        args = dict(
            product_id=self.settings.product_id,
            firmware=self.settings.firmware,
            panel_id=self.settings.panel_id,
            pc_password=encoded_password,
            user_code=0x000000,
            _not_used1=0x19,
            source_id=0x02,
        )

        logger.info("Installer login")
        reply = await self.core.send_wait(
            parsers.InitializeCommunication,
            args=args,
            reply_expected=[0x10, 0x70, 0x00],
        )

        if reply is None:
            logger.error("Installer login failed")
            return False

        if reply.fields.value.po.command == 0x10:
            logger.info("Authentication Success")
            return True
        elif (
            reply.fields.value.po.command == 0x70
            or reply.fields.value.po.command == 0x00
        ):
            logger.error("Authentication Failed. Wrong Password?")
            raise AuthenticationFailed("Wrong PASSWORD")

    @staticmethod
    def _request_status_reply_check(message: Container, address: int):
        mvars = message.fields.value

        if mvars.po.command == 0x5 and mvars.address == address:
            return True

        return False

    async def request_status(self, i: int):
        args = dict(address=self.mem_map["status_base1"] + i)
        reply = await self.core.send_wait(
            parsers.ReadEEPROM,
            args,
            reply_expected=lambda m: self._request_status_reply_check(m, i),
        )
        if reply is not None:
            logger.debug("Received status response: %d" % i)
            return self.handle_status(reply, parsers.RAMDataParserMap)
        else:
            raise StatusRequestException("No reply to status request: %d" % i)

    async def control_zones(self, zones: list, command: str) -> bool:
        """
        Control zones
        :param list zones: a list of zones
        :param str command: textual command
        :return: True if we have at least one success
        """
        if command not in ZONE_ACTIONS:
            return False

        accepted = False

        for zone in zones:
            args = dict(action=ZONE_ACTIONS[command], argument=(zone - 1))
            reply = await self.core.send_wait(
                parsers.PerformAction, args, reply_expected=0x4
            )

            if reply is not None:
                accepted = True

        return accepted

    async def control_partitions(self, partitions: list, command: str) -> bool:
        """
        Control Partitions
        :param list partitions: a list of partitions
        :param str command: textual command
        :return: True if we have at least one success
        """
        if command not in PARTITION_ACTIONS:
            return False

        accepted = False

        for partition in partitions:
            args = dict(action=PARTITION_ACTIONS[command], argument=(partition - 1))
            reply = await self.core.send_wait(
                parsers.PerformAction, args, reply_expected=0x4
            )

            if reply is not None:
                accepted = True

        return accepted

    async def control_outputs(self, outputs, command) -> bool:
        """
        Control PGM
        :param list outputs: a list of pgms
        :param str command: textual command
        :return: True if we have at least one success
        """
        if command not in PGM_ACTIONS:
            return False

        accepted = False

        for output in outputs:
            if command == "pulse":
                args = dict(action=PGM_ACTIONS["on"], argument=(output - 1))
                reply = await self.core.send_wait(
                    parsers.PerformAction, args, reply_expected=0x4
                )
                if reply is None:
                    continue

                await asyncio.sleep(cfg.OUTPUT_PULSE_DURATION)
                args = dict(action=PGM_ACTIONS["off"], argument=(output - 1))
                reply = await self.core.send_wait(
                    parsers.PerformAction, args, reply_expected=0x4
                )
                if reply is not None:
                    accepted = True
            else:
                args = dict(action=PGM_ACTIONS[command], argument=(output - 1))
                reply = await self.core.send_wait(
                    parsers.PerformAction, args, reply_expected=0x4
                )
                if reply is not None:
                    accepted = True

        return accepted

    async def send_panic(self, partitions, panic_type, user_id):
        accepted = False

        args = {"partitions": partitions, "panic_type": panic_type, "user_id": user_id}

        reply = await self.core.send_wait(
            parsers.SendPanicAction, args, reply_expected=0x4
        )

        if reply is not None:
            accepted = True

        return accepted



#!/usr/bin/env python3
import argparse
import binascii
from collections import OrderedDict
import re
import traceback

import yaml

from paradox.connections.ip.parsers import (
    IPMessageCommand,
    IPMessageRequest,
    IPMessageResponse,
    IPMessageType,
    IPPayloadConnectResponse,
)
from paradox.hardware import create_panel
from paradox.hardware.parsers import InitiateCommunicationResponse


class Colors:  # You may need to change color settings
    RED = "\033[31m"
    ENDC = "\033[m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    BLUE = "\033[34m"
    MAGENTA = "\033[95m"
    CYAN = "\033[96m"
    ON_WHITE = "\033[30m\033[47m"


def ordered_load(stream, Loader=yaml.loader.SafeLoader, object_pairs_hook=OrderedDict):
    class OrderedLoader(Loader):
        pass

    def construct_mapping(loader, node):
        loader.flatten_mapping(node)
        return object_pairs_hook(loader.construct_pairs(node))

    OrderedLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, construct_mapping
    )
    return yaml.load(stream, OrderedLoader)


class PayloadParser:
    def __init__(self):
        self.panel = create_panel(None)

    def parse(self, parsed):
        try:
            message_type = parsed.header.message_type
            if message_type == IPMessageType.serial_passthrough_request:
                self._parse_message(parsed.payload, "topanel")
            elif message_type == IPMessageType.serial_passthrough_response:
                self._parse_serial_passthrough_response(parsed)
            elif message_type == IPMessageType.ip_request:
                self._parse_ip_request(parsed)
            elif message_type == IPMessageType.ip_response:
                self._parse_ip_response(parsed)
        except Exception:
            print(f"{Colors.RED}Failed to parse message: \n{parsed}{Colors.ENDC}")
            traceback.print_exc()

    def _parse_ip_response(self, parsed):
        if (
            parsed.header.command == IPMessageCommand.connect
            and parsed.header.sub_command == 0
        ):
            print(
                f"{Colors.ON_WHITE}{IPPayloadConnectResponse.parse(parsed.payload)}{Colors.ENDC}"
            )
            if len(parsed.payload) > 25:
                p = parsed.payload[25:62]
                print(
                    f"{Colors.ON_WHITE}InitiateCommunicationResponse:\n{InitiateCommunicationResponse.parse(p)}{Colors.ENDC}"
                )
        elif parsed.header.command == IPMessageCommand.multicommand:
            self._parse_multicommand(parsed, "frompanel")
        else:
            print(
                f"{Colors.RED}No parser for ip_response payload: {binascii.hexlify(parsed.payload)}{Colors.ENDC}"
            )

    def _parse_multicommand(self, parsed, direction):
        if parsed.header.sub_command == 0x0:  # Generic multicommand
            self._parse_generic_multicommand(parsed, direction)
        elif parsed.header.sub_command == 0x1:  # MGSP multiread
            self._parse_generic_mgsp_multiread(parsed, direction)
        elif parsed.header.sub_command == 0xDD:  # Unified multiread
            pass

    def _parse_serial_passthrough_response(self, parsed):
        parsed_payload = self._parse_message(parsed.payload, direction="frompanel")
        if parsed_payload is not None:
            if parsed_payload.fields.value.po.command == 0:  # panel detection
                self.panel = create_panel(None, parsed_payload)
            if parsed_payload.fields.value.po.command == 5:  # eeprom/ram read
                self._parse_serial_passthrough_eeprom_read(parsed_payload)

    def _parse_serial_passthrough_eeprom_read(self, parsed_payload):
        if (
            "control" in parsed_payload.fields.value
            and parsed_payload.fields.value.control.ram_access
            and parsed_payload.fields.value.control._eeprom_address_bits == 0
            and parsed_payload.fields.value.bus_address == 0
        ):
            ram_address = parsed_payload.fields.value.address
            ram_parser = self.panel.get_message("RAMDataParserMap").get(ram_address)
            if ram_parser is not None:
                print(
                    f"{Colors.ON_WHITE}{ram_parser.parse(parsed_payload.fields.value.data)}{Colors.ENDC}"
                )
            else:
                print(
                    f"{Colors.RED}No parser for {ram_address} ram address, data: {binascii.hexlify(parsed_payload.fields.value.data)}{Colors.ENDC}"
                )

    def _parse_ip_request(self, parsed):
        if parsed.header.command == IPMessageCommand.multicommand:
            self._parse_multicommand(parsed, "topanel")
        else:
            print(
                f"{Colors.RED}No parser for ip_request payload: {binascii.hexlify(parsed.payload)}{Colors.ENDC}"
            )

    def _parse_generic_multicommand(self, parsed, direction: str):
        i = 0
        while i < len(parsed.payload):
            cmd_len = parsed.payload[i]
            i += 1
            cmd = parsed.payload[i : i + cmd_len]
            assert len(cmd) == cmd_len
            i += cmd_len
            print(f"{Colors.ON_WHITE}Multicommand: {cmd}{Colors.ENDC}")

            self._parse_message(cmd, direction)

    def _parse_message(self, message, direction):
        parsed_payload = self.panel.parse_message(message, direction)
        if parsed_payload is not None:
            print(f"{Colors.ON_WHITE}{parsed_payload}{Colors.ENDC}")
        else:
            print(
                f"{Colors.RED}No parser for {direction} message payload: {binascii.hexlify(message)}{Colors.ENDC}"
            )
        return parsed_payload

    def _parse_generic_mgsp_multiread(self, parsed, direction):
        print(f"{Colors.RED}No parser for {direction} mgsp_multiread{Colors.ENDC}")


def old_yaml_format_traverser(data):
    re_key = re.compile(r"peer(\d+)_(\d+)")
    for key, value in data.items():
        # example peer0_0, where 0 is peer number and 0 is index number
        matches = re_key.match(key)

        yield {
            "peer": int(matches.group(1)),
            "data": value,
            "index": int(matches.group(2)),
        }


def decrypt_file(file, password, max_packets: int = None):
    try:
        data = ordered_load(file, yaml.loader.SafeLoader)
        parser = PayloadParser()

        if "peers" in data and "packets" in data:
            iterator = data["packets"]
        else:
            iterator = old_yaml_format_traverser(data)

        n = 0
        for packet in iterator:
            value = packet["data"]
            key = packet["index"]
            if not value[0] == 0xAA:
                print(f"{Colors.RED}Not an IP packet: {value}{Colors.ENDC}")
                continue
            header = value[0:16]
            payload = value[16:]

            is_request = header[3] in [3, 4]
            print(
                f"{Colors.BLUE}PC->IP: " if is_request else f"{Colors.GREEN}IP->PC:\n",
                f"\theader: {binascii.hexlify(header)}\n",
                f"\tencrypted_payload: {binascii.hexlify(payload)}",
            )

            if is_request:
                parsed = IPMessageRequest.parse(value, password=password)
            else:
                parsed = IPMessageResponse.parse(value, password=password)

            if (
                parsed.header.command == IPMessageCommand.connect
                and parsed.header.message_type == IPMessageType.ip_request
            ):
                if parsed.header.sub_command == 0:
                    assert password == parsed.payload, "Wrong decryption password"

            if (
                parsed.header.command == IPMessageCommand.connect
                and parsed.header.message_type == IPMessageType.ip_response
            ):
                if parsed.header.sub_command == 0:
                    password = parsed.payload[1:17]
                    assert len(password) == 16, "Wrong password length"
                    print(f"{Colors.RED}Session password: {password}{Colors.ENDC}")
                elif parsed.header.sub_command == 3:
                    connection_result = parsed.payload[0] & 240
                    if connection_result == 16:
                        print(f"{Colors.RED}Successfully connected{Colors.ENDC}")
                    elif connection_result == 112:
                        print(f"{Colors.RED}Connection failed{Colors.ENDC}")
                    else:
                        print(f"{Colors.RED}Connected to unknown{Colors.ENDC}")

            print(
                f"\tpayload: {binascii.hexlify(parsed.payload)}\n",
                f"\tpayload_raw: {parsed.payload}",
            )

            print(parsed)
            print(Colors.ENDC)
            parser.parse(parsed)

            if not is_request:
                print(
                    "----end %s-------------------------------------------------------------"
                    % key
                )
            n += 1

            if max_packets is not None and n > max_packets:
                print(f"Force stopped on {max_packets} packets")
                return

    except yaml.YAMLError as exc:
        print(f"{Colors.RED}{exc}{Colors.ENDC}")


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "file",
        type=argparse.FileType("r"),
        help="YAML file to parse. In wireshark right click on the first "
        "package, 'Follow->TCP Stream', 'Show and save data as': "
        "'YAML', copy contents to a file.",
    )
    parser.add_argument(
        "password",
        type=str,
        default="paradox",
        help="IP Module password for decryption",
    )
    parser.add_argument(
        "-n",
        "--packets",
        type=int,
        help="Packets to decrypt",
    )

    args = parser.parse_args()

    decrypt_file(args.file, args.password.encode("utf8"), args.packets)


if __name__ == "__main__":
    main()



from paradox.config import config as cfg

ELEMENT_TOPIC_MAP = dict(
    partition=cfg.MQTT_PARTITION_TOPIC,
    zone=cfg.MQTT_ZONE_TOPIC,
    output=cfg.MQTT_OUTPUT_TOPIC,
    pgm=cfg.MQTT_OUTPUT_TOPIC,
    repeater=cfg.MQTT_REPEATER_TOPIC,
    bus=cfg.MQTT_BUS_TOPIC,
    module=cfg.MQTT_MODULE_TOPIC,
    keypad=cfg.MQTT_KEYPAD_TOPIC,
    system=cfg.MQTT_SYSTEM_TOPIC,
    user=cfg.MQTT_USER_TOPIC,
    door=cfg.MQTT_DOOR_TOPIC,
)


def get_control_topic_prefix(element_type):
    return "/".join([
        cfg.MQTT_BASE_TOPIC,
        cfg.MQTT_CONTROL_TOPIC,
        ELEMENT_TOPIC_MAP[element_type],
    ])


def get_state_topic_prefix(element_type):
    return "/".join([
        cfg.MQTT_BASE_TOPIC,
        cfg.MQTT_STATES_TOPIC,
        ELEMENT_TOPIC_MAP[element_type],
    ])


# coding: utf-8

# Adapted from code in https://github.com/konomae/stunpy

import binascii
import random
import socket
import struct

STUN_PORT = 3478

FAMILY_IPv4 = b"\x01"
FAMILY_IPv6 = b"\x02"

# 0b00: Request
# 0b01: Binding
BINDING_REQUEST_SIGN = b"\x00\x01"  # 16bit (2bytes)

BINDING_RESPONSE_ERROR = b"\x01\x11"
BINDING_RESPONSE_SUCCESS = b"\x01\x01"
MAGIC_COOKIE = b"\x21\x12\xA4\x42"  # 32bit (4bytes)

CHANGE_RESPONSE_SUCCESS = b"\x01\x03"

CONNECT_REQUEST = b"\x00\x0a"
CONNECT_RESPONSE_SUCCESS = b"\x01\x0a"

CONNECTION_BIND_REQUEST = b"\x00\x0b"
CONNECTION_BIND_SUCCESS = b"\x01\x0b"

CONNECTION_REFRESH_REQUEST = b"\x00\x04"

# STUN Attribute Registry
MAPPED_ADDRESS = b"\x00\x01"
RESPONSE_ADDRESS = b"\x00\x02"
CHANGE_REQUEST = b"\x00\x03"
SOURCE_ADDRESS = b"\x00\x04"
CHANGED_ADDRESS = b"\x00\x05"
USERNAME = b"\x00\x06"
PASSWORD = b"\x00\x07"
MESSAGE_INTEGRITY = b"\x00\x08"
ERROR_CODE = b"\x00\x09"
UNKNOWN_ATTRIBUTES = b"\x00\x0A"
REFLECTED_FROM = b"\x00\x0B"
REALM = b"\x00\x14"
NONCE = b"\x00\x15"
XOR_MAPPED_ADDRESS = b"\x80\x82"
XOR_PEER_ADDRESS = b"\x00\x12"
REQUESTED_TRANSPORT = b"\x00\x19"
TRANSPORT_TCP = b"\x06"
LIFETIME = b"\x00\x0d"
REQUESTED_ADDRESS_TYPE = b"\x00\x17"
RESERVED3 = b"\x00\x00\x00"
RESERVED1 = b"\x00"
CONNECTION_ID = b"\x00\x2a"

STUN_ATTRIBUTE_NAMES = {
    MAPPED_ADDRESS: "MAPPED-ADDRESS",
    RESPONSE_ADDRESS: "RESPONSE-ADDRESS",
    CHANGE_REQUEST: "CHANGE-REQUEST",
    SOURCE_ADDRESS: "SOURCE-ADDRESS",
    CHANGED_ADDRESS: "CHANGED-ADDRESS",
    USERNAME: "USERNAME",
    PASSWORD: "PASSWORD",
    MESSAGE_INTEGRITY: "MESSAGE-INTEGRITY",
    ERROR_CODE: "ERROR-CODE",
    UNKNOWN_ATTRIBUTES: "UNKNOWN-ATTRIBUTES",
    REFLECTED_FROM: "REFLECTED-FROM",
    REALM: "REALM",
    NONCE: "NONCE",
    XOR_MAPPED_ADDRESS: "XOR-MAPPED-ADDRESS",
    b"\x00\x20": "XOR-MAPPED-ADDRESS",
}


def generate_transaction_id():
    tid = []
    for i in range(24):  # 96bits (12bytes)
        tid.append(random.choice("0123456789ABCDEF"))
    return binascii.a2b_hex("".join(tid))


def build_binding_request(transaction_id):
    if len(transaction_id) != 12:
        raise Exception("Invalid transaction id")

    body_length = b"\x00\x00"
    return b"".join([BINDING_REQUEST_SIGN, body_length, MAGIC_COOKIE, transaction_id])


def build_change_request(transaction_id):
    if len(transaction_id) != 12:
        raise Exception("Invalid transaction id")

    body_length = b"\x00\x18"
    return b"".join(
        [
            CHANGE_REQUEST,
            body_length,
            MAGIC_COOKIE,
            transaction_id,
            REQUESTED_TRANSPORT,
            b"\x00\x04",
            TRANSPORT_TCP,
            RESERVED3,
            LIFETIME,
            b"\x00\x04",
            b"\x00\x00\x02\x58",
            REQUESTED_ADDRESS_TYPE,
            b"\x00\x04",
            FAMILY_IPv4,
            RESERVED3,
        ]
    )


def build_connection_bind_request(transaction_id, connection_id):
    if len(transaction_id) != 12:
        raise Exception("Invalid transaction id")

    if len(connection_id) != 4:
        raise Exception("Invalid connection id {}".format(len(connection_id)))

    body_length = b"\x00\x08"
    return b"".join(
        [
            CONNECTION_BIND_REQUEST,
            body_length,
            MAGIC_COOKIE,
            transaction_id,
            CONNECTION_ID,
            b"\x00\x04",
            connection_id,
        ]
    )


def build_connection_refresh_request(transaction_id):
    if len(transaction_id) != 12:
        raise Exception("Invalid transaction id")

    body_length = b"\x00\x08"
    return b"".join(
        [
            CONNECTION_REFRESH_REQUEST,
            body_length,
            MAGIC_COOKIE,
            transaction_id,
            LIFETIME,
            b"\x00\x04",
            b"\x00\x00\x02\x58",
        ]
    )


def build_connect_request(transaction_id, xoraddr=None):
    if len(transaction_id) != 12:
        raise Exception("Invalid transaction id")

    if xoraddr is None:
        raise Exception("Invalid connect address")

    body_length = b"\x00\x0c"
    return b"".join(
        [
            CONNECT_REQUEST,
            body_length,
            MAGIC_COOKIE,
            transaction_id,
            XOR_PEER_ADDRESS,
            b"\x00\x08",
            xoraddr,
        ]
    )


def validate_response(buf, transaction_id):
    if not buf or len(buf) < 20:
        raise Exception("Response too short")

    response_magic_cookie = buf[4:8]
    if MAGIC_COOKIE != response_magic_cookie:
        raise Exception("Invalid magic cookie")

    response_transaction_id = buf[8:20]
    if transaction_id != response_transaction_id:
        raise Exception("invalid transaction id")


def ip_to_bytes(ip, xor):
    octets = [binascii.a2b_hex("%02x" % int(o)) for o in ip.split(".")]
    addr_int = struct.unpack("!I", b"".join(octets))[0]

    if xor:
        magicCookieBytesInt = int(binascii.b2a_hex(MAGIC_COOKIE), 16)
        addr_int = magicCookieBytesInt ^ addr_int

    addr_bytes = binascii.a2b_hex("%08x" % addr_int)
    return addr_bytes


def port_to_bytes(port, xor):
    if xor:
        magicCookieHighBytesInt = int(binascii.b2a_hex(MAGIC_COOKIE[:2]), 16)
        port = magicCookieHighBytesInt ^ port

    port_bytes = binascii.a2b_hex("%04x" % port)
    return port_bytes


def read_mapped_address(attr_type, attr_body, attr_len):
    assert attr_type in (MAPPED_ADDRESS, XOR_MAPPED_ADDRESS, b"\x00\x20")
    assert attr_body[:1] == b"\x00"

    family_bytes = attr_body[1:2]
    port_bytes = attr_body[2:4]
    addr_bytes = attr_body[4:attr_len]
    xor = attr_type in (XOR_MAPPED_ADDRESS, b"\x00\x20")

    family_text = ""
    assert family_bytes in (FAMILY_IPv4, FAMILY_IPv6)
    if family_bytes == FAMILY_IPv4:
        family_text = "IPv4"
    elif family_bytes == FAMILY_IPv6:
        family_text = "IPv6"

    # TODO: IPv6s
    port = int(binascii.b2a_hex(port_bytes), 16)
    addr_int = int(binascii.b2a_hex(addr_bytes), 16)

    if xor:
        # port
        magicCookieHighBytesInt = int(binascii.b2a_hex(MAGIC_COOKIE[:2]), 16)
        port = magicCookieHighBytesInt ^ port

        # addr
        magicCookieBytesInt = int(binascii.b2a_hex(MAGIC_COOKIE), 16)
        addr_int = magicCookieBytesInt ^ addr_int

    octets = struct.pack("!I", addr_int)
    ip = ".".join([str(c) for c in octets])
    return dict(
        name=STUN_ATTRIBUTE_NAMES[attr_type], ip=ip, port=port, family=family_text
    )


def read_attributes(attributes, body_length):
    pos = 0
    parsed_attributes = []
    while pos < body_length:
        attr_type = attributes[pos : pos + 2]  # 16bit (2bytes)
        attr_len = int(
            binascii.b2a_hex(attributes[pos + 2 : pos + 4]), 16
        )  # 16bit (2bytes)
        attr_body = attributes[pos + 4 : pos + 4 + attr_len]

        if attr_type in (MAPPED_ADDRESS, XOR_MAPPED_ADDRESS, b"\x00\x20"):
            parsed_attributes.append(
                read_mapped_address(attr_type, attr_body, attr_len)
            )
        else:
            if attr_type == b"\x80\x22":
                attr_body = attr_body.decode("utf-8")
            else:
                attr_body = attr_body.hex()
            parsed_attributes.append(
                dict(
                    name=STUN_ATTRIBUTE_NAMES.get(attr_type),
                    attr_type=attr_type.hex(),
                    attr_body=attr_body,
                    attr_len=attr_len,
                )
            )

        attr_head = 2 + 2  # attr_type + attr_len
        remain = attr_len % 4
        padding = 4 - remain if remain else 0
        pos += attr_head + attr_len + padding

    return parsed_attributes


def is_error(body):
    for element in body:
        if element["name"] == "ERROR-CODE":
            return True

    return False


def get_error(body):
    for element in body:
        if element["name"] == "ERROR-CODE":
            body = binascii.unhexlify(element["attr_body"])
            return {"class": body[:3], "code": body[3], "reason": body[4:]}

    return False


class StunClient(object):
    def __init__(self, host, port=STUN_PORT):
        self.sock = None
        self.transaction_id = None
        self.req = None
        self.transaction_id = generate_transaction_id()

        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.sock.bind(("0.0.0.0", 0))
        self.host = host
        self.port = port
        self.sock.connect((host, port))

    def send_tcp_change_request(self):
        self.req = build_change_request(self.transaction_id)
        self.sock.send(self.req)

    def send_binding_request(self):
        self.req = build_binding_request(self.transaction_id)
        self.sock.send(self.req)

    def send_connect_request(self, host=None, port=None, xoraddr=None):
        if host is not None and port is not None:
            xorhost = ip_to_bytes(host, True)
            xorport = port_to_bytes(port, True)
            addr = b"".join([xorport, xorhost])
        elif xoraddr is not None:
            addr = xoraddr
        else:
            raise Exception("Invalid arguments")

        self.req = build_connect_request(self.transaction_id, addr)
        self.sock.send(self.req)

    def send_connection_bind_request(self, connection_id):
        self.req = build_connection_bind_request(self.transaction_id, connection_id)
        self.sock.send(self.req)

    def receive_response(self):
        buf = self.sock.recv(2048)
        validate_response(buf, self.transaction_id)

        body_length = int(binascii.b2a_hex(buf[2:4]), 16)
        attributes = buf[20:]
        assert len(attributes) == body_length

        return read_attributes(attributes, body_length)

    def send_refresh_request(self):
        self.req = build_connection_refresh_request(self.transaction_id)
        self.sock.send(self.req)

    def get_socket(self):
        return self.sock

    def close(self):
        if self.sock:
            self.sock.close()



import datetime
from enum import Enum

from construct import Adapter


class DateAdapter(Adapter):
    def _decode(self, obj, context, path):
        return datetime.datetime(obj[0] * 100 + obj[1], obj[2], obj[3], obj[4], obj[5])

    def _encode(self, obj, context, path):
        return [
            obj.year / 100,
            obj.year % 100,
            obj.month,
            obj.day,
            obj.hour,
            obj.minute,
        ]


class ModuleSerialAdapter(Adapter):
    def _decode(self, obj, context, path):
        return hex(
            int(obj[0]) * 10
            ^ 8 + int(obj[1]) * 10
            ^ 4 + int(obj[2]) * 10
            ^ 2 + int(obj[3]) * 10
            ^ 0
        )


class PartitionStateAdapter(Adapter):
    states = dict(arm=4, disarm=5, arm_sleep=3, arm_stay=1, none=0)

    def _decode(self, obj, context, path):
        for k, v in enumerate(self.states):
            if v == obj[0]:
                return k

        return "unknown"

    def _encode(self, obj, context, path):
        if obj in self.states:
            return self.states[obj]

        return 0


class ZoneStateAdapter(Adapter):
    states = dict(bypass=0x10)

    def _decode(self, obj, context, path):
        for k, v in enumerate(self.states):
            if v == obj[0]:
                return k

        return "unknown"

    def _encode(self, obj, context, path):
        if obj in self.states:
            return self.states[obj]

        return 0


class StatusAdapter(Adapter):
    def _decode(self, obj, context, path):
        r = dict()
        for i in range(0, len(obj)):
            status = obj[i]
            for j in range(0, 8):
                r[i * 8 + j + 1] = ((status >> j) & 0x01) == 0x01

        return r


class PartitionStatusAdapter(Adapter):
    def _decode(self, obj, context, path):
        partition_status = dict()

        for i in range(0, 2):
            partition_status[i + 1] = dict(
                alarm=(obj[0 + i * 4] & 0xF0 != 0)
                or (obj[2 + i * 4] & 0x80 != 0),  # Combined status
                pulse_fire_alarm=obj[0 + i * 4] & 0x80 != 0,
                audible_alarm=obj[0 + i * 4] & 0x40 != 0,
                silent_alarm=obj[0 + i * 4] & 0x20 != 0,
                strobe_alarm=obj[0 + i * 4] & 0x10 != 0,
                arm_stay=obj[0 + i * 4] & 0x04 != 0,
                arm_sleep=obj[0 + i * 4] & 0x02 != 0,
                arm=obj[0 + i * 4] & 0x01 != 0,
                bell_activated=obj[1 + i * 4] & 0x80 != 0,
                auto_arming_engaged=obj[1 + i * 4] & 0x40 != 0,
                recent_closing_delay=obj[1 + i * 4] & 0x20 != 0,
                intellizone_delay=obj[1 + i * 4] & 0x10 != 0,
                zone_bypassed=obj[1 + i * 4] & 0x08 != 0,
                alarms_in_memory=obj[1 + i * 4] & 0x04 != 0,
                entry_delay=obj[1 + i * 4] & 0x02 != 0,
                exit_delay=obj[1 + i * 4] & 0x01 != 0,
                paramedic_alarm=obj[2 + i * 4] & 0x80 != 0,
                _not_used1=obj[2 + i * 4] & 0x40 != 0,
                arm_with_remote=obj[2 + i * 4] & 0x20 != 0,
                transmission_delay_finished=obj[2 + i * 4] & 0x10 != 0,
                bell_delay_finished=obj[2 + i * 4] & 0x08 != 0,
                entry_delay_finished=obj[2 + i * 4] & 0x04 != 0,
                exit_delay_finished=obj[2 + i * 4] & 0x02 != 0,
                intellizone_delay_finished=obj[2 + i * 4] & 0x01 != 0,
                _not_used2=obj[3 + i * 4] & 0x80 != 0,
                wait_window=obj[3 + i * 4] & 0x40 != 0,
                _not_used3=obj[3 + i * 4] & 0x20 != 0,
                in_remote_delay=obj[3 + i * 4] & 0x10 != 0,
                _not_used4=obj[3 + i * 4] & 0x08 != 0,
                stayd_mode_active=obj[3 + i * 4] & 0x04 != 0,
                arm_force=obj[3 + i * 4] & 0x02 != 0,
                ready_status=obj[3 + i * 4] & 0x01 != 0,
            )

        return partition_status


class ZoneStatusAdapter(Adapter):
    def _decode(self, obj, context, path):
        zone_status = dict()
        for i in range(0, len(obj)):
            zone_status[i + 1] = dict(
                was_in_alarm=(obj[i] & 0x80) != 0,
                alarm=(obj[i] & 0x40) != 0,
                fire_delay=(obj[i] & 0b00110000) == 0b00110000,
                entry_delay=(obj[i] & 0b00010000) == 0b00010000,
                intellizone_delay=(obj[i] & 0b00100000) == 0b00010000,
                no_delay=(obj[i] & 0b00110000) == 0,
                bypassed=(obj[i] & 0x08) != 0,
                shutdown=(obj[i] & 0x04) != 0,
                in_tx_delay=(obj[i] & 0x02) != 0,
                was_bypassed=(obj[i] & 0x01) != 0,
            )

        return zone_status


class PGMStatusAdapter(Adapter):
    def _decode(self, obj, context, path):
        pgm_status = dict()
        for i in range(0, len(obj)):
            pgm_status[i + 1] = dict(
                on=(obj[i] & 32) != 0,
            )

        return pgm_status


class SignalStrengthAdapter(Adapter):
    def _decode(self, obj, context, path):
        strength = dict()
        for i in range(0, len(obj)):
            strength[i + 1] = obj[i]
        return strength


class PGMDefinitionAdapter(Adapter):
    class PGMEvent(Enum):
        zone_ok = 0
        zone_open = 1
        partition_status = 2
        bell_status = 3
        non_reportable_event = 6
        remote_control_access = 7
        pgm_activation_8 = 8
        pgm_activation_9 = 9
        pgm_activation_10 = 10
        pgm_activation_11 = 11
        cold_start_wireless_zone = 12
        cold_start_wireless_output_module = 13
        bypass_programming = 14
        user_code_activated_output = 15
        wireless_smoke_maintenance_signal = 16
        delay_zone_alarm_transmission = 17
        zone_signal_strength_weak_1 = 18
        zone_signal_strength_weak_2 = 19
        zone_signal_strength_weak_3 = 20
        zone_signal_strength_weak_4 = 21
        button_pressed_on_remote1 = 22
        button_pressed_on_remote2 = 23
        fire_delay_started = 24
        upload_download_software_access = 26
        bus_module_added_removed = 27
        stayd_pass_acknowledged = 28
        arming_with_user = 29
        special_arming = 30
        disarming_with_user = 31
        disarming_after_alarm_with_user = 32
        alarm_cancelled_with_user = 33
        special_disarming = 34
        zone_bypassed = 35
        zone_in_alarm = 36
        fire_alarm = 37
        zone_alarm_restore = 38
        fire_alarm_restore = 39
        special_alarm = 40
        zone_shutdown = 41
        zone_tampered = 42
        zone_tamper_restored = 43
        new_trouble = 44
        trouble_restored = 45
        bus_module_new_trouble = 46
        bus_module_trouble_restored = 47
        special = 48
        low_battery_on_zone = 49
        low_battery_on_zone_restore = 50
        zone_supervision_trouble = 51
        zone_supervision_restore = 52
        wireless_output_supervision_trouble = 53
        wireless_output_supervision_restore = 54
        wireless_output_tamper_trouble = 55
        wireless_output_tamper_trouble_restore = 56
        non_medical_alarm = 57
        zone_was_forced = 58
        zone_included = 59
        force_zone_has_been_rearmed = 60
        system_status = 64

    def _decode(self, obj, context, path):
        disabled = sum(obj) == 0

        if disabled:
            return "disabled"

        try:
            activation_event = (
                PGMDefinitionAdapter.PGMEvent(obj[0])
                if sum(obj[:3]) != 0
                else "disabled"
            )
        except Exception:
            activation_event = "unknown"

        try:
            deactivation_event = (
                PGMDefinitionAdapter.PGMEvent(obj[3])
                if sum(obj[3:]) != 0
                else "disabled"
            )
        except Exception:
            deactivation_event = "unknown"

        return dict(activation=activation_event, deactivation=deactivation_event)



import codecs
import logging
import re
from importlib import import_module

logger = logging.getLogger("PAI").getChild(__name__)

codec_cache = {}


def paradox_codec_search(name: str):
    name = name.lower()
    match = re.match("^paradox[-_]([a-z]{2})$", name)
    if match:
        enc = match.group(1)
        mod = codec_cache.get(enc)
        if mod is not None:
            return mod
        else:
            try:
                mod = codec_cache[enc] = getregentry(enc)
                logger.debug(f"Loaded {name} encoding")
                return mod
            except ImportError:
                pass         


def getregentry(enc):

    module = import_module("." + enc, "paradox.lib.encodings.charmaps")
    
    decoding_table = module.charmap
    encoding_table = codecs.charmap_build(decoding_table)


    class Codec(codecs.Codec):
        def encode(self, input, errors="strict"):
            return codecs.charmap_encode(input, errors, encoding_table)

        def decode(self, input, errors="strict"):
            return codecs.charmap_decode(input, errors, decoding_table)


    class IncrementalEncoder(codecs.IncrementalEncoder):
        def encode(self, input, final=False):
            return codecs.charmap_encode(input, self.errors, encoding_table)[0]


    class IncrementalDecoder(codecs.IncrementalDecoder):
        def decode(self, input, final=False):
            return codecs.charmap_decode(input, self.errors, decoding_table)[0]


    class StreamWriter(Codec, codecs.StreamWriter):
        pass


    class StreamReader(Codec, codecs.StreamReader):
        pass

    return codecs.CodecInfo(
        name="paradox-"+enc,
        encode=Codec().encode,
        decode=Codec().decode,
        incrementalencoder=IncrementalEncoder,
        incrementaldecoder=IncrementalDecoder,
        streamreader=StreamReader,
        streamwriter=StreamWriter,
    )


def register_encodings():
    codecs.register(paradox_codec_search)



import logging

logger = logging.getLogger("PAI").getChild(__name__)

MODULES = {
    "construct": dict(
        mandatory=True, desc="basic operation", install_name="construct>=2.9.43"
    ),
    "argparse": dict(
        mandatory=True, desc="basic operation", install_name="argparse>=1.4.0"
    ),
    "slugify": dict(
        mandatory=True,
        desc="basic operation (remove non-ascii from mqtt topic names)",
        install_name="python-slugify>=4.0.1",
    ),
    "pyserial": dict(
        mandatory=False, desc="the serial connection", install_name="pyserial>=3.4"
    ),
    "serial_asyncio": dict(
        mandatory=False,
        desc="the serial connection",
        install_name="pyserial-asyncio>=0.4",
    ),
    "pushbullet": dict(
        mandatory=False,
        desc="the Pushbullet interface",
        install_name="pushbullet.py=>0.11.0",
    ),
    "requests": dict(
        mandatory=False, desc="the IP150 connection", install_name="requests>=2.20.0"
    ),
    "ws4py": dict(
        mandatory=False, desc="the Pushbullet interface", install_name="ws4py>=0.4.2"
    ),
    "yaml": dict(
        mandatory=False, desc="the IP150 connection", install_name="pyyaml>=5.2.0"
    ),
    "pydbus": dict(
        mandatory=False, desc="the Signal interface", install_name="pydbus>=0.6.0"
    ),
    "pygobject": dict(
        mandatory=False, desc="the Signal interface", install_name="pygobject>=3.20.0"
    ),
    "gi": dict(mandatory=False, desc="the Signal interface", install_name="gi>=1.2"),
    "pytz": dict(
        mandatory=False, desc="Panel time sync", install_name="pytz>=2023.3.post1"
    ),
    "mqtt": dict(mandatory=False, desc="MQTT", install_name="paho_mqtt>=1.5.0,<2"),
    "pre-commit": dict(mandatory=False, desc="Development", install_name="pre-commit"),
    "flake8": dict(mandatory=False, desc="Code checker", install_name="flake8"),
    "tox": dict(mandatory=False, desc="virtual env / testing", install_name="tox"),
}


def import_error_help(error):
    logger.error(f"Could not import Python3 module '{error.name}': {error}\n")

    if error.name in MODULES:
        m = MODULES[error.name]
        logger.error(
            "The module is required for {} and IS{} mandatory.".format(
                m["desc"], " NOT" if not m["mandatory"] else ""
            )
        )
        if not m["mandatory"]:
            logger.error(
                "If you do not require such functionality, you can disable it in the config file,"
            )
            logger.error("and skip installing the module.\n")
        logger.error("To install ONLY this module, execute: \n")
        logger.error(" pip3 install '{}'\n".format(m["install_name"]))

    logger.error(
        "To install ALL modules required, go to the main project folder and execute:\n"
    )
    logger.error(" pip3 install -r requirements.txt\n")
    logger.error(
        "Take in consideration that the 'requirements.txt' file only has the most common modules enabled."
    )
    logger.error("Uncomment the entries as required for your setup.")
    logger.error(
        "ATTENTION: If this module is not listed there, please report the bug."
    )
    import sys

    sys.exit(-1)



from paradox.interfaces.mqtt.entities.abstract_entity import AbstractControllableEntity
from paradox.lib.utils import sanitize_key


class Switch(AbstractControllableEntity):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.hass_entity_type = "switch"

    def serialize(self):
        config = super().serialize()
        config.update(dict(
            state_on="True",
            state_off="False",
        ))
        return config


class ZoneBypassSwitch(Switch):
    def __init__(self, zone, device, availability_topic: str):
        super().__init__(device, availability_topic)
        self.zone = zone

        self.key = sanitize_key(zone["key"])
        self.label = zone["label"]
        self.property = "bypassed"

        self.pai_entity_type = "zone"

    def serialize(self):
        config = super().serialize()
        config.update(dict(
            payload_on="bypass",
            payload_off="clear_bypass",
        ))
        return config


class PGMSwitch(Switch):
    def __init__(self, pgm, device, availability_topic: str):
        super().__init__(device, availability_topic)
        self.pgm = pgm

        self.key = sanitize_key(pgm["key"])
        self.label = pgm["label"]
        self.property = "on"

        self.pai_entity_type = "pgm"



import logging
from collections import defaultdict
from typing import Dict

from construct import Container

from paradox.data.element_type_container import ElementTypeContainer

logger = logging.getLogger("PAI").getChild(__name__)


def _iterate_properties(data):
    if isinstance(data, list):
        for key, value in enumerate(data):
            yield (key, value)
    elif isinstance(data, dict):
        for key, value in data.items():
            if type(key) == str and key.startswith("_"):  # ignore private properties
                continue
            yield (key, value)


def convert_raw_status(raw_status: Container) -> Dict[str, ElementTypeContainer]:
    out = defaultdict(ElementTypeContainer)

    for root_key, value in _iterate_properties(raw_status):
        element_type = root_key.split("_")[0]
        prop_name = "_".join(root_key.split("_")[1:])

        if not prop_name:
            prop_name = element_type

        out[element_type].deep_merge(_parse_raw_status(prop_name, value))

    return out


def _parse_raw_status(prop_name: str, value) -> (dict, list):
    if isinstance(value, dict):
        arr = {}
        all_keys_are_int = all([isinstance(k, int) for k in value.keys()])
        for element_key, element_value in _iterate_properties(value):
            if (
                all_keys_are_int
            ):  # we deal with int keys. it should be a list of elements
                arr[element_key] = _parse_raw_status(prop_name, element_value)
            else:  # we deal with text keys
                if isinstance(
                    element_value, dict
                ):  # ignore prop_name, go deeper to find keys
                    arr[element_key] = _parse_raw_status(element_key, element_value)
                else:  # use element_key instead of prop_name
                    arr[element_key] = element_value
        return arr
    else:
        return {prop_name: value}



from paradox.interfaces.mqtt.entities.alarm_control_panel import AlarmControlPanel
from paradox.interfaces.mqtt.entities.binary_sensors import ZoneStatusBinarySensor, \
    SystemBinarySensor, PartitionBinarySensor
from paradox.interfaces.mqtt.entities.sensor import PAIStatusSensor, SystemStatusSensor, ZoneNumericSensor
from paradox.interfaces.mqtt.entities.switch import ZoneBypassSwitch, PGMSwitch


class MQTTAutodiscoveryEntityFactory:
    def __init__(self, availability_topic, device=None):
        self.availability_topic = availability_topic
        self.device = device

    def set_device(self, device):
        self.device = device

    def make_alarm_control_panel_config(self, partition, code: str = None):
        return AlarmControlPanel(partition, self.device, self.availability_topic, code)

    def make_partition_status_binary_sensor(self, partition, status):
        return PartitionBinarySensor(partition, status, self.device, self.availability_topic)

    def make_pai_status_sensor(self, pai_status_topic):
        return PAIStatusSensor(pai_status_topic, self.device, self.availability_topic)

    def make_zone_bypass_switch(self, zone):
        return ZoneBypassSwitch(zone, self.device, self.availability_topic)

    def make_zone_status_binary_sensor(self, zone, status):
        return ZoneStatusBinarySensor(zone, status, self.device, self.availability_topic)

    def make_zone_status_numeric_sensor(self, zone, status):
        return ZoneNumericSensor(zone, status, self.device, self.availability_topic)

    def make_pgm_switch(self, pgm):
        return PGMSwitch(pgm, self.device, self.availability_topic)

    def make_system_status(self, system_key, status):
        if system_key == 'troubles':
            return SystemBinarySensor(system_key, status, self.device, self.availability_topic)
        else:
            return SystemStatusSensor(system_key, status, self.device, self.availability_topic)



from unittest.mock import AsyncMock

import pytest

from paradox.hardware import Panel
from paradox.paradox import Paradox


@pytest.mark.asyncio
async def test_send_panic(mocker):
    alarm = Paradox()
    alarm.panel = mocker.Mock(spec=Panel)
    alarm.panel.send_panic = AsyncMock()

    alarm.storage.get_container("partition").deep_merge(
        {1: {"id": 1, "key": "Partition 1"}}
    )
    alarm.storage.get_container("user").deep_merge({3: {"id": 3, "key": "User 3"}})

    assert await alarm.send_panic("1", "fire", "3")
    alarm.panel.send_panic.assert_called_once_with([1], "fire", 3)
    alarm.panel.send_panic.reset_mock()

    assert await alarm.send_panic("Partition 1", "fire", "User 3")
    alarm.panel.send_panic.assert_called_once_with([1], "fire", 3)


@pytest.mark.asyncio
async def test_control_doors(mocker):
    alarm = Paradox()
    alarm.panel = mocker.Mock(spec=Panel)
    alarm.panel.control_doors = AsyncMock()

    alarm.storage.get_container("door").deep_merge({1: {"id": 1, "key": "Door 1"}})

    assert await alarm.control_door("1", "unlock")
    alarm.panel.control_doors.assert_called_once_with([1], "unlock")
    alarm.panel.control_doors.reset_mock()

    assert await alarm.control_door("Door 1", "unlock")
    alarm.panel.control_doors.assert_called_once_with([1], "unlock")



import asyncio
import logging

from gi.repository import GLib
from pydbus import SystemBus

from paradox.config import config as cfg
from paradox.event import EventLevel, Notification

# Signal interface.
# Only exposes critical status changes and accepts commands
from paradox.interfaces.text.core import ConfiguredAbstractTextInterface
from paradox.lib import ps

logger = logging.getLogger("PAI").getChild(__name__)


class SignalTextInterface(ConfiguredAbstractTextInterface):
    """Interface Class using Signal"""

    def __init__(self, alarm):
        super().__init__(
            alarm,
            cfg.SIGNAL_EVENT_FILTERS,
            cfg.SIGNAL_ALLOW_EVENTS,
            cfg.SIGNAL_IGNORE_EVENTS,
            cfg.SIGNAL_MIN_EVENT_LEVEL,
        )

        self.signal = None
        self.glib_loop = None

    def stop(self):
        """Stops the Signal Interface Thread"""
        if self.glib_loop is not None:
            self.glib_loop.quit()

        super().stop()

        logger.debug("Signal Stopped")

    async def run(self):
        await super().run()

        bus = SystemBus()

        self.signal = bus.get("org.asamk.Signal")
        self.signal.onMessageReceived = self.handle_message
        self.glib_loop = GLib.MainLoop()

        logger.debug("Signal Interface Running")

        asyncio.get_event_loop().run_in_executor(None, self.glib_loop.run)

    def send_message(self, message: str, level: EventLevel):
        if self.signal is None:
            logger.warning("Signal not available when sending message")
            return

        try:
            self.signal.sendMessage(str(message), [], cfg.SIGNAL_CONTACTS)
        except Exception:
            logger.exception("Signal send message")

    def handle_message(self, timestamp, source, groupID, message, attachments):
        """Handle Signal message. It should be a command"""

        logger.debug(
            "Received Message {} {} {} {} {}".format(
                timestamp, message, groupID, message, attachments
            )
        )

        if source in cfg.SIGNAL_CONTACTS:
            future = asyncio.run_coroutine_threadsafe(
                self.handle_command(message), self.alarm.work_loop
            )
            ret = future.result(10)

            m = f"Signal {source} : {ret}"
            logger.info(m)
        else:
            m = f"Signal {source} (UNK): {message}"
            logger.warning(m)

        self.send_message(m, EventLevel.INFO)
        ps.sendNotification(
            Notification(sender=self.name, message=m, level=EventLevel.INFO)
        )



from paradox.data.model import DetectedPanel


class Device:
    def __init__(self, panel: DetectedPanel):
        self.panel = panel

    @property
    def serial_number(self):
        return self.panel.serial_number

    @property
    def model(self):
        return self.panel.model

    @property
    def firmware_version(self):
        return self.panel.firmware_version

    def serialize(self):
        return dict(
            manufacturer="Paradox",
            model=self.model,
            identifiers=[f"Paradox_{self.model}_{self.serial_number}"],
            name=self.model,
            sw_version=self.firmware_version,
        )